// cluster.cc: clustering
//
// Copyright (C) 2000 Bruce I Mills <B.I.Mills@cs.cf.ac.uk>
// Copyright (C) 2001-2002 Frank C Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.

#include "config.h"
#include "points.hxx"
#include "cluster.hxx"
#include "realisation.hxx"

// ================== Heap class needed for sorting ==================

// This is an implimentation of a priority queue 

heap::heap (dualgraph const &g, pointset const &ps)
{
  n_max = g.n + 1;
  key = new numeric [n_max];
  item = new word [n_max];
  n = 0;
  word l = 0;
  while (l < g.n) {
    push ((ps.item[g.item[l].vertex[0]] - 
	   ps.item[g.item[l].vertex[1]]) . length(), l);
    ++l;
  }
}

void heap::put (FILE *f) const
{
  fprintf(f,"{\n");
  for (word i = 0; i < n; ++i)
    fprintf (f, "  (%f : %d)\n", key[i], item[i]);
  fprintf(f,"}\n");
}

void heap::empty()
{
  n = 0;
}

void heap::push (numeric k, word p)
{
  if (n >= n_max)
    return;
  key[n] = k;
  item[n] = p;
  up ();
}

word heap::grab ()
{
  if (n > 0) {
    word p = item[0];
    down ();
    return p;
  }
  return 0;
}

// This routine throws away the top element in the heap
void heap::down ()
{
  --n;
  key[0] = key[n];
  item[0] = item[n];

  word where = 0;
  while (right_child (where) < n) {

    word dc = key[left_child (where)] <= key[right_child (where)] ? 
      left_child (where) : right_child (where);

    if (key[where] < key[dc])
      return;
    
    numeric tk = key[where];
    key[where] = key[dc];
    key[dc] = tk;
    
    word ti = item[where];
    item[where] = item[dc];
    item[dc] = ti;
    
    where = dc;
  }

  if (left_child (where) < n) {
    word dc = left_child (where);
    
    if (key[where] < key[dc])
      return;
    
    numeric tk = key[where];
    key[where] = key[dc];
    key[dc] = tk;
    
    word ti = item[where];
    item[where] = item[dc];
    item[dc] = ti;
    
    where = dc;
  }

  return;
}

// This routine adds in the element just outside the valid heap
void heap::up ()
{
  // Note: since parent(0) = 0 eventually the condition is false

  word where = n;
  while (key[where] < key[parent(where)]) {
    numeric tk = key[where];
    key[where] = key[parent (where)];
    key[parent (where)] = tk;
    
    word ti = item[where];
    item[where] = item[parent (where)];
    item[parent (where)] = ti;
    
    where = parent (where);
  }

  ++n;

  return;
}



// ================= partition class ============================


// This class is a generic partition on the set of integers
// from 0 up to n-1. The idea is that it can be used to partition
// any indexed object, by partitioning the indices.
// However this partition also keeps track of some information
// as it goes along, such as how many nodes and links in each class
// and how many classes. The number of classes and the number of
// nodes in a class is clearly defined. By the number of links 
// what is meant is the number of merges that has been done to
// produce the given class. This information is specifically
// designed to help in examination of connected components of
// graphs.

// Create the finest partition possible on the given number of
// indices.  Ie, the partition with k classes, each containing exactly
// one index.

partition::partition (word k)
{
  parent = new word[k];
  depth = new word[k];
  links = new word[k];
  nodes = new word[k];
  for (n = 0; n < k; ++n) {
    parent[n] = n;
    depth[n] = 0;
    links[n] = 0;
    nodes[n] = 1;
  }
  incomplete = 0;
  classes = k;
}

// In this partition, merge the classes indicated by the given
// indices.
void partition::merge(word a, word b)
{
  if (a >= n || b >= n) {
    cerr << progname << ": error in merge\n";
    exit (1);
  }

  // Get class root
  word x = root(a);
  word y = root(b);

  if (x == y) {
    incomplete -= (nodes[x] * (nodes[x] - 1) != 2 * links[x]);
    ++links[x];
    incomplete += (nodes[x] * (nodes[x] - 1) != 2 * links[x]);
    return;
  }

  incomplete -= (nodes[y] * (nodes[y] - 1) != 2 * links[y]);
  incomplete -= (nodes[x] * (nodes[x] - 1) != 2 * links[x]);
  --classes;

  if(depth[x] > depth[y]) {
    parent[x] = y; 
    nodes[y] += nodes[x];
    links[y] += links[x] + 1;
    incomplete += (nodes[y] * (nodes[y] - 1) != 2 * links[y]);
  } else if(depth[y] > depth[x]) {
    parent[y] = x; 
    nodes[x] += nodes[y];
    links[x] += links[y] + 1;
    incomplete += (nodes[x] * (nodes[x] - 1) != 2 * links[x]);
  } else {
    ++depth[x];
    parent[y] = x;
    nodes[x] += nodes[y];
    links[x] += links[y] + 1;
    incomplete += (nodes[x] * (nodes[x] - 1) != 2 * links[x]);
  }
  return;
}

// Factor the pointset according to the information in this partition.
// Ie, for each class of indices, collect the points refered to and 
// compute the centroid. Then output the set of centroids. This means
// that the resulting pointset has exactly one point for each implicit
// class in the original pointset. Thus replacing each group of points
// in the pointset with the centroid of the group. The basic intention
// is to create the pointset up to a given tolerance. The partition 
// will typically have been constructed on the basis of two points being
// in the same class if and only if their positions are within tolerance 
// of each other. 

pointset *partition::factor (pointset *p) const
{
  pointset *q = new pointset (classes, p->item[0].d);
  word k = 0;
  for (word j = 0; j < n; ++j) {
    if (j == root (j)) {
      point sum = origen (p->item[0].d);
      word count = 0;
      for (word i = 0; i < n; ++i) {
	if (j == root(i)) {
	  sum += p->item[i];
	  ++count;
	}
      }
      q->item[k++] = sum / count;
    }
  }
  return q;
}

// Factor the given dualgraph according to this partition.
//
// In the simplest sense this means that we just equate all the
// appropriate vertices. That is, given a particular class we
// replace each vertex index in that class with a single new
// label. One pragmatic way of doing this is to replace each
// index label with the root of the class that it belongs to.
//
// However, in the more involved sense we should decide what 
// to do to factor the rest of the structure. The first matter
// would be that any two edges that now join the same two vertices
// should be equated (since this is an assumption of some of the
// regularity analysis algorithms). However, doing this can produce
// an edge that joins more than two faces, which prevents the
// structure from being a dual graph. At the time of writing this
// comment I have not implemented any solution to this, although
// I do have some ideas about how to go about it.
//
// This is also the point at which stitching problems should be fixed.
// ACIS models can have a mis-match between the adjacency information
// as implied by the geometry, and that specified in the combinatorics.
// In particular, two faces might abut each other, like pieces of paper
// that have been moved into edge contact, and yet this information is
// not included in the combinatorial part of the model. In this case 
// the problem is fixed in principle by equating vertices and edges 
// that are in close contact with each other. If this is the only
// problem with the model, then the direct approach here will create
// the natural combinatorial model for the geometry. (This is assuming
// of course that the combinatorial structure of adjacent faces matches).

dualgraph *partition::factor (dualgraph *dg) const
{
  dualgraph *ng = new dualgraph (*dg);
  for (word i = 0; i < ng->n; ++i) {
    ng->item[i].vertex[0] = root (ng->item[i].vertex[0]);
    ng->item[i].vertex[1] = root (ng->item[i].vertex[1]);
  }
  return ng;
}

word partition::put (FILE *f, char *s1, char *s2)
{
  word count = 0;
  if(s1)
    count += fprintf (f, "%s", s1);

  count += fprintf (f, "{");

  for (word j = 0; j < n; ++j) {
    if (j == root (j)) {
      count += fprintf (f, "{");
      bool first = true;
      for (word i = 0; i < n; ++i) {
	if (j == root(i)) {
	  if (!first) {
	    count += fprintf (f, " %d", i);
	  } else {
	    first = false;
	    count += fprintf (f, "%d", i);
	  }
	}
      }
      count += fprintf (f, "}");
    }
  }
  count += fprintf(f,"}\n");

  if (s2)
    count += fprintf (f, "%s", s2);

  return count;
}

// Find the root of the current class, 
// this is really a purely internal, implimentation issue.
word partition::root(word a) const
{
  if (a >= n)
    return n;

  word b = a;
  while (b != parent[b])
    b = parent[b];

  return b;
}
