// create.cc: create special point sets
//
// Copyright (C) 2000 Bruce I Mills <B.I.Mills@cs.cf.ac.uk>
// Copyright (C) 2001-2002 Frank C Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.

#include "config.h"
#include "xfig.hxx"
#include "points.hxx"

// Pi
#ifndef PI
#define PI 3.14159265358979323846
#endif // PI

// These are a couple of routines intended for the construction of
// reasonably regular pointsets for testing the algorithms. The basic
// idea is to treat the points as charged particles, and then let them
// find a lowest energy configuration. It works reasonably well, in
// that it produces an irregularly oriented approximate version of
// things such as cubes, and octahedrons. Which is basically what is
// wanted. However it is not so good with a larger number of
// points. Could be improved, but it is really just a side issue, I
// have copies of various appropriate data files already. Keep this in
// mind though for generating some rough and ready data sets with
// interesting properties.  Especially for things such as partial
// symmetries.

// Try to create an icosahedron.

numeric energy (pointset *p)
{
  numeric t = 0.0;
  for (word i= 0; i < p->n; ++i) {
    for (word j = i + 1; j < p->n; ++j) {
      numeric a = (p->item[i] - p->item[j]) . length ();
      t += 1.0 / (a * a + 0.00000001);
    }
  }
  return t;
}

numeric improve (pointset *p, numeric delta)
{
  numeric t = energy (p);
  for(word i = 0; i<p->n; ++i) {

    for (long l0 = -1L; l0 < 2L; ++l0) {
      for (long l1 = -1L; l1 < 2L; ++l1) {
	for (long l2 = -1L; l2 < 2L; ++l2) {
	  point x = p->item[i];
	  p->item[i].coord[0] += static_cast<numeric> (l0) * delta;
	  p->item[i].coord[1] += static_cast<numeric> (l1) * delta;
	  p->item[i].coord[2] += static_cast<numeric> (l2) * delta;
	  p->item[i] = p->item[i] / p->item[i].length();
	  numeric a = energy(p);
	  if (a < t)
	    t = a;
	  else
	    p->item[i] = x; 
	}
      }
    }

  }

  return energy (p);
}

pointset *spread (word n)
{
  pointset *p = new pointset (n, 3);
  for (word i = 0; i < p->n; ++i)
    p->item[i].set3 (1.0, 0.0, 0.0);

  numeric delta = 0.1;
  numeric t = energy (p);
  for (word i = 0; i < 5000 && delta > 0.0000001; ++i) {
    numeric e = improve (p, 0.004);
    if (t == e)
      delta *= 0.99;
    else
      t = e;
#ifdef DEBUG
    if (!(i % 100))
      cerr << "energy = " << t << ", delta = " << delta << endl;
#endif // DEBUG
  }

  return p;
}

// This function generates a set of points that describe a prism.  The
// actual points form `along' number of polygons, of `around' numnber
// of vertices.

pointset *prism (word around, word along)
{
  word num = along * around;
  pointset *ps = new pointset (num, 3);
  for (word j = 0; j < along; ++j) {
    for (word i = 0; i < around; ++i) {
      ps->item[j * around + i] .
	set3 (100.0 * cos ((2.0 * PI / static_cast<numeric> (around)) * 
			   static_cast<numeric> (i)),
	      100.0 * sin ((2.0 * PI / static_cast<numeric> (around)) * 
			   static_cast<numeric> (i)),
	      10.0 * static_cast<numeric> (j));
    }
  }
  return ps;
}
