// freg.cc: main
//
// Copyright (C) 2000 Bruce I Mills <B.I.Mills@cs.cf.ac.uk>
// Copyright (C) 2001-2002 Frank C Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.

#include "config.h"
#include "sil_getopt.h"
#include "points.hxx"
#include "realisation.hxx"
#include "build.hxx"
#include "create.hxx"

// Numerical precision
numeric epsilon;
char *progname;

// This is the entry point for the code. The basic function that main
// actually performs is the natural function for main, that is, the
// parsing and interpretation of the command line. There are actually
// quite a few actions that freg can perform. Some indication of these
// is given at the start of the main routine, via the printout that is
// produced for the -h flag. Basically, freg can read in and process
// geometrical data as an xfig file (possibly created by xfig), and in
// the freg format (using the freg type text formats discussed in the
// section on types in this file), and produce an freg file with a set
// of partially regular points in it for testing purposes. By default
// it reads an freg file from stdin and determines the symmetries.
// With a single argument treats it as the name of an freg file, opens
// it, and determines the symmetries.

int main (int argc, char **argv)
{
  // Defaults
  epsilon = DEFAULT_EPSILON;
  numeric multiply = -1.0;
  bool usage = false;
  bool error = false;
  bool distance = false;
  bool first_only = false;
  bool rectify = false;
  int makeset = 0;
  progname = argv[0];

  // Getopt options
  static struct sil_option long_cmdline_options[] = {
    {"help",                no_argument,       NULL, 'h'},
    {"version",             no_argument,       NULL, 'v'},
    {"no-cluster",          no_argument,       NULL, 'n'},
    {"rectify",             no_argument,       NULL, 'r'},
    {"distance-matrix",     no_argument,       NULL, 'm'},
    {"multiply",            required_argument, NULL, 'd'},
    {"make-prism-pointset", required_argument, NULL, 'p'},
    {"make-sym-pointset",   required_argument, NULL, 's'},
    {"epsilon",             required_argument, NULL, 'e'},
    {NULL,                  no_argument,       NULL, 0}
  };

  while (true) {
    int c = sil_getopt_long (argc, argv, "hvnrmd:p:s:e:",
			     long_cmdline_options, NULL);
    if (c == -1)
      break;
    switch (c) {
    case 'h':
      usage = true;
      break;
    case 'v':
      cout << "FReg (" << argv[0] << ") Version 0.2 (" << TARGET_TYPE << ")\n"
	   << "Copyright (C) 2000 Bruce I Mills <B.I.Mills@cs.cf.ac.uk>\nCopyright (C) 2001-2002 Frank C Langbein <frank@langbein.org>\nAll rights reserved. See the GNU General Public Licence Version 2 or later.\n";
      exit (0);
      break;
    case 'n':
      first_only = true;
      break;
    case 'r':
      rectify = true;
      break;
    case 'm':
      distance = true;
      break;
    case 'e':
      if (sscanf (sil_optarg, "%lg", &epsilon) != 1) {
	cerr << argv[0] << ": illegal argument for -e";
	exit (1);
      }
      break;
    case 'd':
      if (sscanf (sil_optarg, "%lg", &multiply) != 1 || multiply < epsilon) {
	cerr << argv[0] << ": illegal argument for -d";
	exit (1);
      }
      break;
    case 's':
      if (sscanf (sil_optarg, "%d", &makeset) != 1 || makeset < 1) {
	cerr << argv[0] << ": illegal argument for -m";
	exit (1);
      }
      break;
    case 'p':
      if (sscanf (sil_optarg, "%d", &makeset) != 1 || makeset < 1) {
	cerr << argv[0] << ": illegal argument for -p";
	exit (1);
      }
      makeset = -makeset;
      break;
    default:
      // Illegal option/argument.
      error = true;
    }
  }

  char *file = NULL;
  if (sil_optind < argc)
    file = argv[sil_optind++];
  if (sil_optind < argc) {
    cerr << argv[0] << ": too many file names\n";
    exit (1);
  }

  if (!error &&
      ((distance && (multiply > 0.0 || makeset != 0 || rectify || first_only)) ||
       (multiply > 0.0 && (distance || makeset != 0 || rectify || first_only)) ||
       (makeset != 0 && (distance || multiply > 0.0))))
    error = true;

  if (error) {
    cerr << argv[0] << ": illegal arguments\n"
	 << argv[0] << ": type `" << argv[0] << " -h' for help\n";
    exit (1);
  }

  if (usage) {
    // print out the usage information
    cout << "Usage: " << argv[0] << " [<options>] <file>\n"
	 << "  <file>         Input/output filename in freg or xfig format\n"
	 << " Options:\n"
	 << "   --help|-h                      Print usage information\n"
	 << "   --version|-v                   Print version information\n"
	 << "   --no-cluster|-n                Do not cluster points for symmetry detection\n"
	 << "   --rectify|-r                   Rectify points after symmetry detection\n"
	 << "   --distance-matrix|-m           Generate distance matrix for pointset\n"
	 << "   --multiply|-d <float>          Multiply coordinates with <float>\n"
	 << "   --make-sym-pointset|-s <num>   Generate point set by minimising energy\n"
	 << "   --make-prism-pointset|-p <num> Generate prism point set\n";
    exit (0);
  }

  if (makeset != 0) {
    // Generate point set by minimizing energy or as prism
    pointset *ps = (makeset > 0) ? spread (makeset) : prism (-makeset, 2);
    ps->put (true, false, file);
    delete ps;
    exit (0);
  }

  if (multiply > 0.0) {
    // Expand pointset as indicated by multiply.
    pointset *ps = new pointset (file);
    (*ps) *= multiply;
    ps->put (true);
    delete ps;
    exit (0);
  }
  
  if (distance) {
    // Generate the distance matrix for a pointset
    pointset *ps = new pointset (file);
    ps->put (true);
    ps->metric ();
    delete ps;
    exit (0);
  }

  realisation *r = new realisation (file);

  // Find symmetries
  find_factored_regularities (stdout, r, rectify, first_only);

  delete r;

  exit (0);
}
