// maps.cc: symmetry mappings
//
// Copyright (C) 2000 Bruce I Mills <B.I.Mills@cs.cf.ac.uk>
// Copyright (C) 2001-2002 Frank C Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.

#include "config.h"
#include "xfig.hxx"
#include "points.hxx"
#include "realisation.hxx"
#include "cluster.hxx"
#include "maps.hxx"

// ================= injection class ============================

// Create identity injections
void injection::identity(word m, word w)
{
  word imin = m;
  word imax = m + w;
  
  if (imax > n_max) {
    cerr << progname << ": injection size overflow\n";
    exit (1);
  }
  
  n = w;
  
  for (word i = 0; i < imin; ++i)
    forward[i] = reverse[i] = n_max;
  
  for (word i = imin; i < imax; ++i)
    forward[i] = reverse[i] = i;
  
  for(word i = imax; i < n_max; ++i)
    forward[i] = reverse[i] = n_max;

  return;
}

// Figure out whether this is a permutation or not.
// The basic idea is to find the number of ends of chains,
// ie something that is in the domain, but not the range,
// or vice versa. If there are none, then this is a permutation.
// This operation has to look at the entire injection array. It 
// could be done much faster by updating a count of chain ends
// each time a pair is added to the permutation. However, at the
// time of writing this routine is only going to be called in order
// to print the injection out in the correct format. So it is not
// going to cause any major inefficiency. But the mod is there to
// be done if you want call this lots of times as part of some 
// processing or other.

bool injection::is_perm() const
{
  for (word i = 0; i < n; ++i)
    if((forward[i] < n_max) ^ (reverse[i] < n_max))
      return false;
  return true;
}

// Force the pair (x,y) into the injection.
// If either of x or y are in the range then one or two other
// pairs may have to be deleted in order to maintain the 
// injective nature of the set of pairs. The size requires
// a bit of fiddling to maintain. It may be incremented or
// decremented or remain unchanged. If y is not in the range,
// then the pair (x,b) is deleted if it exists. Similarly for
// x not being in the domain.

void injection::addpair (word x, word y)
{
  if (x >= n_max || y >= n_max) {
    cerr << progname << ": injection size overflow\n";
    exit (1);
  }
  
  if (forward[x] == y && reverse[y] == x) 
    return;
  
  if (forward[x] == n_max && reverse[y] == n_max) {
    ++n;
    forward[x] = y;
    reverse[y] = x;
    return;
  }
  
  if (forward[x] != n_max) {
    reverse[forward[x]] = n_max;
    --n;
  }
  if (reverse[y] != n_max) {
    forward[reverse[y]] = n_max;
    --n;
  }
  forward[x]=y;
  reverse[y]=x;
  ++n;

  return;
}

bool injection::cando (realisation *r, word min, word max) const
{
  for (word i = min; i <= max; ++i)
    if (r->face_type (i) != r->face_type(forward[i]))
      return false;
  return true;
}

bool injection::cando (dualgraph *dg, injection *edge_perm,
		       injection *face_perm)
{
  // get the original edge list
  word old_edges[dg->n * 3];
  for (word i = 0; i < dg->n; ++i) {
    old_edges[i * 3    ] = dg->item[i].vertex[0];  // 1st edge
    old_edges[i * 3 + 1] = dg->item[i].vertex[1];  // 2nd edge
    old_edges[i * 3 + 2] = i;                      // record number
  }
  // and put it into standard form
  fudge_sort (3, dg->n, 0, 1, old_edges);

  // apply the injection to get the transformed edge list
  word new_edges[dg->n * 3];
  for (word i = 0; i < dg->n; ++i) {
    new_edges[i * 3    ] = forward[dg->item[i].vertex[0]];  // 1st edge
    new_edges[i * 3 + 1] = forward[dg->item[i].vertex[1]];  // 2nd edge
    new_edges[i * 3 + 2] = i;                              // record number
  }
  // and put it into standard form
  fudge_sort (3, dg->n, 0, 1, new_edges);

  // To determine if this vertex permutation induces an edge
  // permutation, just include all the required pairs of edges and
  // check to see if the result is an injection of the correct size.
  // If not, then this vertex permutation is not compatible with the
  // given dualgraph. Note, by applying the vertex permutation to the
  // new edges, and then sorting, any specific record, r, in the new
  // edge list is moved to the location of the record that it has been
  // changed into. Thus the mapping is from the new_edge record number
  // to the old_edge record number.
  edge_perm->identity ();
  for (word i = 0; i < dg->n; i++)
    edge_perm->addpair (new_edges[i * 3 + 2], old_edges[i * 3 + 2]);
  if (edge_perm->n != dg->n)
    return false;

  // At this point we know that this vertex permutation induces
  // an edge permutation, and we know what that permutation is.
  // The idea is to find out if there is also a face permutation.
  
  // get a copy of the edges in the face graph, as they stand
  word old_faces[dg->n * 3];
  for (word i = 0; i < dg->n; ++i) {
    old_faces[i * 3    ] = dg->item[i].face[0];  // 1st edge
    old_faces[i * 3 + 1] = dg->item[i].face[1];  // 2nd edge
    old_faces[i * 3 + 2] = i;                    // record number
  }
  
  // get the transformed collection of edges
  word new_faces[dg->n * 3];
  for (word i = 0; i < dg->n; ++i) {
    new_faces[i * 3    ] = dg->item[edge_perm->forward[i]].face[0]; // 1st edge
    new_faces[i * 3 + 1] = dg->item[edge_perm->forward[i]].face[1]; // 2nd edge
    new_faces[i * 3 + 2] = i;                                // record number
  }

  // Now, conceptually, this is the oposite of the situation with the
  // vertex graph. There we had a node (vertex) permutation, which meant 
  // we could determine immediately the one place that an edge would have
  // to go in a compatible edge permutation. However, here we have an
  // edge permutation, and we have to find out wether there is a node
  // (face this time), permutation that is compatible. If we just try to
  // match the two lists above we find that there are two ways that each
  // line could be matched, which in a direct implementation would mean
  // 2^n operations. So we have to use more information.
  //
  // Conceptually, we could proceed as follows. In any connected component
  // of the face graph when we find out that edge (a,b) is mapped to edge
  // (c,d) then we know that the node permutation must map a->c and b->d,
  // or a->d and b->c. There are only two options. Now once we make a 
  // decision as to which way to do it, we check out an edge such as 
  // (b,e), which must be mapped to (d,x) for some node x. Further we
  // know that e->x. So by doing a search on the face graph we can 
  // build up the entire node permutation. If the first choice of mapping
  // the first edge does not work, then we can try the second choice. 
  //
  // However, our data structure do not allow a straight graph search, 
  // since what we have is a list of edges. Instead what we can do is
  // start by making a choice about the first edge on the list and then
  // running down the list matching up the other node of all edges in 
  // which exactly one node is already matched, checking each time that 
  // this does not cause a contradition. We need to run down the list
  // at most a number of times equal to the depth of the graph. So, 
  // since the permutation is stored to allow constant time look up
  // in both directions, the number of operations is of the order 
  // (number of edges)x(depth of graph). Which is no more (and typicaly
  // a lot less than) the product of the number of nodes and the 
  // number of edges.

  // first try a straight match
  face_perm->identity ();
  face_perm->addpair (old_faces[0], new_faces[0]);
  face_perm->addpair (old_faces[1], new_faces[1]);
  
  bool finished = fudge_extend (face_perm, old_faces, new_faces, dg->n);

  if (!finished) {
    // if not, then try swaping them over
    face_perm->identity ();
    face_perm->addpair (old_faces[0], new_faces[1]);
    face_perm->addpair (old_faces[1], new_faces[0]);
    finished = fudge_extend (face_perm, old_faces, new_faces, dg->n);
  }

  return finished;
}

// keeping the same domain and range, go to the next
// injection, lexicographically speaking.
//
// Imagine the sequence laid out from left to right.
// Starting at the right go left until you find a
// number that is smaller than the one to the right.
// Call this Ni, find the smallest number bigger than
// Ni, to the right of Ni, and swap it with Ni, now
// reverse all the numbers to the right of posn i.
// 
// Try it for a small number of items ...
// 
// 1234    .... start like this
// 1243    .... swap last pair
// 1324    .... now we need to use three digits
// 1324    .... swap last pait
// 
// and so on. Each time we basically make the smallest number
// that we have not yet used. So this order has some neat properties.

void injection::next ()
{
  // find ascending sequence
  word i;
  for (i = n - 2; (i > 0) && (forward[i] > forward[i + 1]); --i);
  
  // if it is the whole permutation then stop
  if (forward[i] > forward[i + 1])
    return;

  // other wise find pivot to the right of first descent
  word j;
  for (j = n - 1; forward[j] < forward[i]; --j);
  
  // swap first descent with pivot
  word t = forward[i];
  forward[i] = forward[j];
  forward[j] = t;
  
  // now reverse the rest of the sequence
  for(j = n - 1, i++; i < j; ++i, --j) {
    word t = forward[i];
    forward[i] = forward[j];
    forward[j] = t;
  }

  return;
}

// Write an injection to a file in a human understandable format.
// The basic idea is that the injection is composed of a collection
// of chains, eg 1->2->3->4 and loops 1->2->3->1, we represent a
// a chain as (1,2,3,4) and a loop (1 2 3) with the implicit mapping
// of the last element to the first.

word injection::put(FILE *f, char *s1, char *s2)
{
  word count = 0;
  if (s1)
    count += fprintf (f, "%s", s1);

  byte mark[n_max];
  for (word i = 0; i < n_max; ++i)
    mark[i] = 0;

  count += fprintf (f, "|");
  for (word i = 0; i < n_max; ++i) {
    if (!(mark[i]&1)) {
      if (forward[i] < n_max) {
	word a = i; // find start of chain
	// mark bit 1 for each chain element
	while (reverse[a] < n_max && !(mark[a]&2)) {
	  a = reverse[a];
	  mark[a] |= 2;
	} 
	// start at minimum index if the chain is a loop (asthetic choice)
	char sep = ',';
	if (reverse[a] < n_max) {
	  sep=' ';
	  a = forward[a];
	}
	count += fprintf (f, "(");
	for (; a < n_max && !(mark[a]&1); a = forward[a]) {
	  count += fprintf (f, "%d", a);
	  mark[a] |= 1;
	  if (!(mark[forward[a]]&1) && forward[a] < n_max) 
	    count += fprintf (f, "%c", sep);
	}
	count += fprintf (f, ")");
      }
    }

  }

  count += fprintf (f, "|");
  if (s2)
    count += fprintf (f, "%s", s2);

  return count;
}

// This comment is for the `cando' routine, that is a fair 
// way down below, since this comment is fairly lengthy.
//
// Work out if this injection can be consistently applied to
// the given dualgraph. That is, see if the injection, treated
// as a map on the collection of points, induces compatible
// behaviour in terms of edges and faces. In principle this
// is done by checking that pairs of points that are connected
// by an edge are mapped to other pairs of points that are also
// connected by an edge, and that collections of points that are
// on the boundary of any given face are mapped to some collection
// of points on the boundary of a face.
//
// On the presumption that there will be a relatively small number
// of edges in a typical dualgraph, at least compared to the square
// of the number of points, we base the first step on points. It 
// also has the advantage that we can use the information gained
// in checking the faces.
//
// The information about which points are on the boundary of which
// edges is stored directly in the dualgraph. Ie, the boundary of
// edge number n is the nth pair of points on the list of edges 
// for the vertex graph. The boundary of face n is the list of 
// indices of edges in the face graph that have n as either a
// first or second element.
//
// If the injection is a complete permutation, and no pair of
// vertices are joined by more than one edge, then if we apply
// the permutation to the indices as listed in the vertex graph,
// we must produce a permutation of the list. If we apply said
// permutation to the face graph list, then we must get a list
// which unifies (see universal algebra) with the original list.
// If and only if these two things occur is the permutation 
// compatible with the combinatorial information stored in the
// dualgraph.
//
// We can check that a permutation is generated on the vertex
// graph by sorting the list of pairs of vertices, and then 
// running through them replacing the vertex names, and then
// sorting it again. If we end up with the same list, then 
// we have a permutation. In order to figure out what the
// permutation is, we could attach the index of the pair of
// vertices to it while it is being sorted, and then just
// match up the indices in the two lists. This gets n log n
// rather than n^2.
//
// We now apply this same permutation to the face graph, and
// check to see if it unifies with the original. This is actually
// just a linear scan through, since there are no axioms to be
// considered. Just attempt to build up a list of the matching
// between the names, ie, a permutation of the face numbers,
// if you try to put two entries in the same slot in the array,
// then this is not unifiable, and if you get all the way through,
// then it is unifiable, and you have the matching.
//
// So this shows us how to determine in O(n log n) whether the
// original permutation of points is compatible with the 
// dualgraph, and what the induced edge and face permutations
// are, as long as we know that each pair of vertices are 
// joined by at most one edge.
//
// Ok, now what if there are pairs of vertices that are joined by 
// multiple edges? In the general case of a dualgraph, a fast solution 
// to the problem involves a fast solution to the graph isomorphism 
// problem. Firstly, we consider making a face and edge version of an 
// edge and vertex structure. Ie, just promote each edge to a face, 
// each vertex to an edge. For example, (like an extrusion), a triangle 
// becomes a triangular prism of infinite length (no nodes). Now squeeze 
// it off at two spots down to a single point, creating two vertices, 
// and keep only the bit inbetween the two points, creating an analogue 
// of a banana structure. Now, we start with two graphs, and create
// the banana version, and place them in space so that the four
// points involved are at the corners of a square. Now in order
// to find if the permutations of the points (that preserve
// distance) also are compatible with the whole structure we
// are going to have to find out if the two bananas are 
// isomorphic. But the information about the points is of
// no use in doing this. We need basically to check for isomorphism
// from a standing start, and this involves solving the original
// graph isomorphism problem.
//
// Of course if we include the assumption that the whole structure
// is embedable in space, no faces passing through each other, then
// the graph has to be embedable in a plane. Which would simplify
// things a lot, if we had the embedding. In principle we could get
// this from the model, and maybe that is a way to go, but it involves
// a lot of mucking about. Further, the complex examples of this will
// not occur in objects for reverse engineering in the context of
// this project. So, for example, for objects in which the surface
// is a manifold, the ambiguity must be limited to having a collection
// of faces that are connected around in a circle. So, the graph 
// isomorphism problem would just be one of checking that two loops
// are isomorphic, a much easier task. 
//
// At the time of writing I have left out the question of this
// ambiguity. Essentially the code below assumes that there are
// no two edges joining the same pair of vertices. Technically
// the algorithm is a bit more robust than this, since in some 
// structures (such as a teseract scheme embedded in three space)
// are such that no matter what choice you make, you get essentially
// the same result. However, this is an unusual situation.

void injection::print_fudge (word width, word height, word *data)
{
  for(word i = 0; i < width * height; ++i) {
    printf ("%d ", data[i]);
    if (!((i + 1) % width))
      printf("\n");
  }
  printf("\n");
  return;
}

bool injection::fudge_extend (injection *face_perm, word *old_faces,
			      word *new_faces, word n)
{
  bool mod;

  do {

    mod = false;

    for (word i = 0; i < n; ++i) {

      word a = old_faces[3 * i];
      word b = old_faces[3 * i + 1];
      word x = new_faces[3 * i];
      word y = new_faces[3 * i + 1];
    
      // look at mapping {a,b} to {x,y}
      bool da = face_perm->domain (a);
      bool db = face_perm->domain (b);
      bool dx = face_perm->range (x);
      bool dy = face_perm->range (y);
      
      if (da || db || dx || dy) {
	word ma = face_perm->forward[a];
	word mb = face_perm->forward[b];
	word mx = face_perm->reverse[x];
	word my = face_perm->reverse[y];
	// in order for the pair to be compatible,
	// the same number of entries on either side 
	// must be already mapped, if not, then there
	// is no permutation possible
	//
	// if exactly one on each side is mapped, check and maybe extend
	if (!da &&  db && !dx &&  dy) {
	  if (mb != y || my != b)
	    return false;
	  mod = true;
	  face_perm->addpair (a, x);
	} else if (!da &&  db &&  dx && !dy) {
	  if (mb != x || mx != b)
	    return false;
	  mod = true;
	  face_perm->addpair (a, y);
	} else if ( da && !db && !dx &&  dy) {
	  if (ma != y || my != a)
	    return false;
	  mod = true;
	  face_perm->addpair (b, x);
	} else if ( da && !db &&  dx && !dy) {
	  if (ma != x || mx != a)
	    return false;
	  mod = true;
	  face_perm->addpair (b, y);
	} else if ( da &&  db &&  dx &&  dy) {
	  // if both are mapped on each side, just check the matching
	  if ((ma != x || mb != y || mx != a || my != b) &&
	      (ma != y || mb != x || mx != b || my != a))
	    return false;
	}
      }
    }

  } while (mod);

  return true;
}


// The reason that we want to do a sort is to get the two lists
// of edges into a standard form. So we want an ordering that
// works on the actual edges, rather than depending on which 
// vertex happened to be first. One obvious way is to order
// according to the maximum index first, and then the minimum.

byte injection::fudge_test(word a1, word a2, word b1, word b2)
{
  word maxa = (a1 > a2) ? a1 : a2;
  word mina = (a1 > a2) ? a2 : a1;
  word maxb = (b1 > b2) ? b1 : b2;
  word minb = (b1 > b2) ? b2 : b1;

  if (maxa > maxb)
    return 1;
  if (maxa < maxb)
    return 0;
  if (mina > minb)
    return 1;
  if (mina < minb)
    return 0;

  return 0;
}

void injection::fudge_sort (word width, word height, word which1, word which2, 
			    word *data)
{
  word done;
  do {
    done = 1;
    for (word i = 0; i < height - 1; ++i) {
      if( fudge_test(data[(i    ) * width + which1],
		     data[(i    ) * width + which2],
		     data[(i + 1) * width + which1],
		     data[(i + 1) * width + which2])) {
	done=0;
	for (word j = 0; j < width; ++j) {
	  word temp = data[i * width + j];
	  data[i * width + j] = data[(i + 1) * width + j];
	  data[(i + 1) * width + j] = temp;
	}
      }
    }
  } while(!done);

  return;
}

// ===================== pseudosym class ========================

// This type stores the permutation information about a representation,
// on the assumption that we have already decided on the partitioning of
// the object, the tolerance, and the particular sub representation that
// we will be working on. Technically, the sub representation is implicit
// in the occurence of symbols in the collection of permutations. But, the
// tolerance and partitioning obviously have to be stored seperately.
//
// In direct terms, this is just a list of permutations. Any regular 
// structure that it has, such as being a group, must be established
// independently. In practical terms, in the code as written, this is
// really just a place to put the collection of permutations that have
// been developed for an object. In principle it could be incorporated
// into a fuller approximate partial symmetry type, which also included
// the tolerance and partitioning, and maybe the sub realisation as well.


// Print this pseuodsym to a file as indicated, in a form intended
// to be reasonably readable by a human. The basic format is each
// permutation is on a single line, followed by a comment indicating
// its index in the list of permutations. The format of the permutation
// is the product of cycles notation. This notation has the problem that
// its textual length varies depending on the number of cycles. However,
// this length must be less than the greatest case, which is for the
// identity permutation (0)(1)(2) ..., which takes twice the number of
// elements in the permutation worth of characters for the brackets,
// and then the number of digits required to write the numbers. Given
// that these numbers are 0,1,2 ..., then the text width is a piecewise
// affine function of the number of items in the permutation. The particular
// calculation used in this routine will underestimate the length if the
// numbers are not contiguous from 0. However, the estimate is intended as
// a rough sort of idea that will tend to line things up, rather than
// attempting to be exactly right every time.

word pseudosym::put (FILE *f, char *s1, char *s2)
{
  word count = 0;

  word est = 0;
  if (n > 0) {
    int m = item[0]->n;
    if (m < 10)
      est = 3 * m - 0 * (  1) + 3;
    else  if (m < 100)
      est = 4 * m - 1 * ( 10) + 4;
    else if (m < 1000)
      est = 5 * m - 2 * (100) + 5;
  }

  if(s1)
    count += fprintf (f, "%s", s1);

  count += fprintf (f, "{\n");
  for (word i = 0; i < n; ++i) {
    count += fprintf (f, "  ");
    word c;
    count += (c = item[i]->put (f));
    while (c++ < est)
      fprintf (f, " "); // attempt to line up comments
    printf (" # permutation %d \n", i);
  }

  count += fprintf (f, "}\n");
  if(s2)
    count += fprintf (f, "%s", s2);

  return count;
}

void pseudosym::expand ()
{
  n_max += n_inc;
  injection **new_item = new injection*[n_max];
  
  for (word l = 0; l < n; ++l)
    new_item[l] = item[l];

  delete[] item;
  item = new_item;

  return;
}
