// maps.hxx: symmetry mappings
//
// Copyright (C) 2000 Bruce I Mills <B.I.Mills@cs.cf.ac.uk>
// Copyright (C) 2001-2002 Frank C Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.

// ================= injection class ============================

// This class represents a set of ordered pairs that is an injection
// from its domain to its range when treated as a combinatorial map.
// The pragmatics of this class see it being used to store information
// about partial injections between supersets of the injections domain
// and range. 
//
// The basic philosophy for this class is that two static arrays of
// words are kept. One stores the forward version of the injection,
// and the other stores the reverse. In this way it is possible to
// get and modify forward and reverse information in constant time
// operations. It is taken as axiomatic that on contemporary computers
// the provision of a few tens of kilobytes of memory is not an issue.
//
// A pair (x,y) is stored by setting forward[x]=y and reverse[y]=x.
// Indices that are not in the domain are indicated by placing an
// invalid value, specifically, injection_size, into the corresponding
// point in the forward array. Similarly for the range.
//
// One way that is used to write injections in a text format is the
// set-of-ordered-pairs notation. Eg, {(1,2),(2,4)} for the injection
// that maps 1 to 2 and 2 to 4. This can be written with a few less
// characters as (1 2)(2 4). This approach can be said to write down
// all the ordered pairs of the injection. Now we could instead write
// down (1 2 4), which is a sequence of elements each that is mapped
// to the next, so 1 is mapped to 2, which is mapped to 4. This notation
// is said to record the open chains of the injection. However when we
// have a permutation, which is a specific type of injection, it can
// become a bit clumsy, since each time the first element is repeated
// as the last. eg (1 2 3 4 1). So intead we can write the closed chains,
// or loops, that form the injection. Eg (1 2 3 4) means that 1 is mapped
// to 2 which is mapped to 3 which is mapped to 4 which is mapped back to
// the first element 1. Now, notice that pairs is really a special case 
// of open chaining, so given the open chain notation, the pairs notation 
// works automatically as a syntactic restriction. 
//
// So, the injection syntax includes open and closed chains. An open
// chain, such as a pair (1,2) or a triple (1,2,3) is denoted by the
// existance of commas between the numbers. A closed chain such as
// (1) or (1 2 3 4), is denoted by spaces between the numbers. Notice
// that an open chain with only one element is not a valid construct,
// so there is no ambiguity here.

class injection
{
public:
  word n_max;      // map size
  word n;          // number of pairs in the map
  word *forward;   // forward map
  word *reverse;   // reverse map
  
  injection (word l) {
    n_max = l;
    n = 0;
    forward = new word[l];
    reverse = new word[l];
  }

  injection (injection const *in) {
    n_max = in->n_max;
    n = in->n;
    forward = new word[n_max];
    reverse = new word[n_max];
    for (word l = 0; l < n_max; ++l) {
      forward[l] = in->forward[l];
      reverse[l] = in->reverse[l];
    }
  }

  ~injection () {
    delete[] forward;
    delete[] reverse;
  }

  injection const &operator= (injection const &in) {
    delete[] forward;
    delete[] reverse;
    n_max = in.n_max;
    n = in.n;
    forward = new word[n_max];
    reverse = new word[n_max];
    for (word l = 0; l < n_max; ++l) {
      forward[l] = in.forward[l];
      reverse[l] = in.reverse[l];
    }
    return *this;
  }

  // make this the identity map on [0 .. w-1]
  void identity (word w) {
    identity (0, w);
  }
  // make this the identity map on [m .. m+w-1]
  void identity (word m = 0, word w = 0);

  // is this the empty map?
  bool is_empty () const {
    return n == 0;
  }
  // is this a permautation?
  bool is_perm () const;

  void addpair (word x, word y);
  
  // is x in the domain
  bool domain (word x) const {
    return x < n_max && forward[x] < n_max;
  }
  // is y in the range
  bool range (word y) const {
    return y < n_max && reverse[y] < n_max;
  }

  // Consistent with face types
  bool cando (realisation *r, word min, word max) const;
  // is this injection consistant with this graph?
  bool cando (dualgraph *dg, injection *edge_perm, injection *face_perm); 

  void next();

  word put(FILE *f = stdout, char *s1 = NULL, char *s2 = NULL);

private:

  void print_fudge (word width, word height, word *data);
  bool fudge_extend (injection *face_perm, word *old_faces, word *new_faces,
		     word n);
  byte fudge_test (word a1, word a2, word b1, word b2);
  void fudge_sort (word width, word height, word which1, word which2, 
		   word *data);

};

// ===================== pseudosym class ========================

// This type stores the permutation information about a representation,
// on the assumption that we have already decided on the partitioning of
// the object, the tolerance, and the particular sub representation that
// we will be working on. Technically, the sub representation is implicit
// in the occurence of symbols in the collection of permutations. But, the
// tolerance and partitioning obviously have to be stored seperately.
//
// In direct terms, this is just a list of permutations. Any regular 
// structure that it has, such as being a group, must be established
// independently. In practical terms, in the code as written, this is
// really just a place to put the collection of permutations that have
// been developed for an object. In principle it could be incorporated
// into a fuller approximate partial symmetry type, which also included
// the tolerance and partitioning, and maybe the sub realisation as well.

class pseudosym
{
public:

  pseudosym (word l) {
    n_max = l;
    n_inc = l;
    n = 0;
    item = new injection*[n_max];
  }

  ~pseudosym () {
    for (word l = 0; l < n; ++l)
      delete item[l];
    delete[] item;
  }

  void add_permutation (injection *p) {
    if (n >= n_max)
      expand ();
    item[n++] = p;
  }
  // Set this structure to the empty set of permutations
  void empty () {
    for (word l = 0; l < n; ++l)
      delete item[l];
    n = 0;
  }
  void rectify (pointset *ps, pointset *newps, point const &c);

  word put (FILE *f = stdout, char *s1 = NULL, char *s2 = NULL);

private:

  void expand ();

  word n_max; // size
  word n_inc; // size increase when running out of space
  word n;     // how many permutations are stored here
  injection **item; // pointer to injections

};
