// points.cc: point class implementation
//
// Copyright (C) 2000 Bruce I Mills <B.I.Mills@cs.cf.ac.uk>
// Copyright (C) 2001-2002 Frank C Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.

#include "config.h"
#include "list.hxx"
#include "xfig.hxx"
#include "points.hxx"

// A class to represent a geometric point

byte point::default_dimension = 3;

// determine the distance of this point from the origen
numeric point::length () const
{
  numeric x = 0.0;
  word l = 0;
  while (l < d) {
    x += coord[l] * coord[l];
    ++l;
  }
  return sqrt (x);
}

point const &point::operator= (point const &p) 
{
  if (d != p.d) {
    delete[] coord;
    d = p.d;
    coord = new numeric[d];
  }
  for (byte l = 0; l < d; ++l)
    coord[l] = p.coord[l];
  return *this;
}

point const &point::operator*= (numeric m)
{
  word l = 0;
  while (l < d)
    coord[l++] *= m;
  return *this;
}

point const &point::operator+= (point const &p1)
{
  if (p1.d != d) {
    if (coord) {
      delete[] coord;
      coord = NULL;
      d = 0;
    }
  } else {
    word l = 0;
    while (l < d) {
      coord[l] += p1.coord[l];
      ++l;
    }
  }
  return *this;
}

point const &point::operator-= (point const &p1)
{
  if (p1.d != d) {
    if (coord) {
      delete[] coord;
      coord = NULL;
      d = 0;
    }
  } else {
    word l = 0;
    while (l < d) {
      coord[l] -= p1.coord[l];
      ++l;
    }
  }
  return *this;
}

// Get point
word point::get (FILE *fp)
{
  char buffer[512];
  word count = 0;
  SimpleList<numeric> cs;
  bool num = false;
  char c;
  while ((c = fgetc (fp)) != EOF) {
    ++count;
    if (num) {
      if (isdigit (c) || c == '.' || c == 'e' || c == 'E' || c == '+' ||
	  c == '-') {
	word l = 0;
	do {
	  buffer[l++] = c;
	  if (l > 511)
	    break;
	  c = fgetc (fp);
	  ++count;
	} while (isdigit (c) || c == '.' || c == 'e' || c == 'E' ||
		 c == '+' || c == '-');
	buffer[l] = '\0';
	ungetc (c, fp);
	--count;
	cs.add_last (new numeric(atof (buffer)));
      } else if (c == ')') {
	if (d != cs.length ()) {
	  if (coord)
	    delete[] coord;
	  d = cs.length ();
	  if (d)
	    coord = new numeric[d];
	}
	word l = 0;
	numeric *r = cs.first ();
	while (r) {
	  coord[l++] = *r;
	  r = cs.next ();
	}
	return count;
      } else if (!isblank (c) && c != ',' && c != ';' && c != '\n') {
	break;
      }
    } else {
      if (c == '(') {
	num = true;
      } else if (!isblank (c)) {
	break;
      }
    }
  }
  cerr << progname << ": error in input file (point format)\n";
  exit (1);
  return 0;
}

// Write a point to a file, by default write it to stdout 
// The extra (char *) arguments are mainly there for convenience, 
// when writing to a the screen, or a file to be read by a person, 
// then it is easier to put in an extra '\n' argument 
// than use printf or putchar. 
// Returns the number of characters actually written.

word point::put (FILE *f, char *s1, char *s2) const
{
  word count = 0;
  if (s1)
    count += fprintf (f, "%s", s1);
  count += fprintf (f, "(");
  if (d) {
    count += fprintf (f, "%4.2f", coord[0]);
    word l = 1;
    while (l < d)
      count += fprintf (f, " %4.2f", coord[l++]);
  }
  count += fprintf (f, ")");
  if (s2)
    count += fprintf (f, "%s", s2);
  return count;
}

// Write the point to a string, works the same way as put
// except that there is no default string to write to so
// the first argument is always the string buffer.

word point::sput(char *f, char *s1, char *s2) const
{
  word count = 0;
  if (s1)
    count += sprintf (f + count, "%s", s1);
  count += sprintf (f + count, "(");
  if (d) {
    count += sprintf (f + count, "%4.2f", coord[0]);
    word l = 1;
    while (l < d)
      count += sprintf (f + count, " %4.2f", coord[l++]);
  }
  count += sprintf (f + count, ")");
  if (s2)
    count += sprintf (f + count, "%s", s2);
  return count;
}

// Some point class operators

// Produce the origen of the coordinate system.
//
// Rather than assume that the origin has all zero co-ordinates
// it is a little cleaner to have an explicit call to get the
// origin. This makes the intention in the code a little clearer
// at times.
point origen (byte n)
{
  point p (n);
  word l = 0;
  while (l < n)
    p.coord[l++] = 0.0;
  return p;
}

// a unit vector in the same direction as this vector based at the origen
point signum (point const &p)
{
  numeric x = p.length ();
  if (x < epsilon)
    return origen (0);
  return p / x;
}

// sum of two points, as vectors based at the origin
point operator+ (point const &p1, point const &p2)
{
  if (p1.d != p2.d)
    return origen (0);
  point p (p1.d);
  word l = 0;
  while (l < p1.d) {
    p.coord[l] = p1.coord[l] + p2.coord[l];
    ++l;
  }
  return p;
}

// difference of two points, as vectors based at the origen
point operator- (point const &p1, point const &p2)
{
  if (p1.d != p2.d)
    return origen(0);
  point p (p1.d);
  word l = 0;
  while (l < p1.d) {
    p.coord[l] = p1.coord[l] - p2.coord[l];
    ++l;
  }
  return p;
}

// divide a point by a scalar, as a vector based at the origen
point operator/ (point const &p1, numeric x)
{
  point p (p1.d);
  word l = 0;
  while (l < p1.d) {
    p.coord[l] = p1.coord[l] / x;
    ++l;
  }
  return p;
}

// multiply a point by a scalar, as a vector based at the origen
point operator*(point const &p1, numeric x)
{
  point p;
  word l = 0;
  while (l < p1.d) {
    p.coord[l] = p1.coord[l] * x;
    ++l;
  }
  return p;
}

// compute the standard inner product, as vectors based at the origen
numeric operator% (point const &p1, point const &p2)
{
  numeric d = 0.0;
  word l = 0;
  while (l < p1.d) {
    d += p1.coord[l] * p2.coord[l];
    ++l;
  }
  return d;
}

// The pointset class definition

pointset::pointset (char *file)
{
  reg_idx = NULL;

  FILE *fp;
  if (file) {
    fp = fopen (file, "r");
    if (!fp) {
      cerr << progname << ": cannot read file " << file << endl;
      exit (1);
    }
  } else {
    fp = stdin;
  }

  char buffer[512];
  fgets (buffer, 511, fp);
  if (strncmp (buffer, "# FReg-1.0", 10) != 0) {
    cerr << progname << ": unkown format, cannot read non-freg file\n";
    fclose (fp);
    exit (1);
  }

  get (fp);

  fclose (fp);
}

// the minimum value obtained by the co-ordinate with the specified index
numeric pointset::cmin (int c) const
{
  if (n == 0)
    return 0.0;
  numeric x = item[0].coord[c];
  for (word i = 1; i < n; ++i) {
    if (x > item[i].coord[c])
      x = item[i].coord[c];
  }
  return x;
}

// the maximum value obtained by the co-ordinate with the specified index

numeric pointset::cmax (int c) const
{
  if (n == 0)
    return 0.0;
  numeric x = item[0].coord[c];
  for (word i = 1; i < n; ++i) {
    if (x < item[i].coord[c])
      x = item[i].coord[c];
  }
  return x;
}

// the width of the range of values obtained by the specified co-ordinate
numeric pointset::cwid (int c) const
{
  return cmax (c) - cmin (c);
}

numeric pointset::character() const
{
  numeric t = (item[0] - item[1]) . length();
  
  for (word i = 0; i < n; ++i) {
    for (word j = i + 1; j < n; ++j) {
      numeric d = (item[i] - item[j]) . length();
      if (t > d)
	t = d;
    }
  }
  
  return t;
}

// the centroid of the pointset

point pointset::centroid () const
{
  point c = origen (n > 0 ? item[0].d : 0);

  word l = 0;
  while (l < n)
    c += item[l++];

  return c / n;
}

// In principle this routine finds the length, area, volume, and so on 
// of a the pointset taken as a simplex. If the number of points is 
// more than the dimension of the space plus one, then the simplex
// must be degenerate, so we just return 0. 
//
// FIXME: only fully implimented up to four points.
//
// I was thinking of using the exterior calculus to generate 
// a general version of this, but in the implimentation envisaged
// only three dimensional raw data was going to be used. So I don't
// need to handle explicitely more than four points. If you feel 
// like it, you can impliment the general thing.

// Compute measure of first m points pus point with index k and centroid c

numeric pointset::measure (word m, word k, point const &c) const
{
  // Only handles a few cases

  // This is with respect to the registration set marked by reg_idx
  // and not the first m points with the additional point k and the
  // centroid c.

  // Length
  if (m == 0)
    return (item[k] - c) . length ();

  // Area
  if (m == 1) {
    point p1 = item[reg_idx[0]] - c;
    point p2 = item[k] - c;
    numeric z0 = p1.coord[1] * p2.coord[2] - p1.coord[2] * p2.coord[1];
    numeric z1 = p1.coord[2] * p2.coord[0] - p1.coord[0] * p2.coord[2];
    numeric z2 = p1.coord[0] * p2.coord[1] - p1.coord[1] * p2.coord[0];
    return sqrt (z0 * z0 + z1 * z1 + z2 * z2);
  }

  // Volume
  if (m == 2) {
    point p1 = item[reg_idx[0]] - c;
    point p2 = item[reg_idx[1]] - c;
    point p3 = item[k] - c;
    return sqr (+ p1.coord[0] * p2.coord[1] * p3.coord[2]
		+ p1.coord[1] * p2.coord[2] * p3.coord[0]
		+ p1.coord[2] * p2.coord[0] * p3.coord[1]
		- p1.coord[2] * p2.coord[1] * p3.coord[0] 
		- p1.coord[0] * p2.coord[2] * p3.coord[1]
		- p1.coord[1] * p2.coord[0] * p3.coord[2]);
  }

  return 0.0;
}

pointset const &pointset::operator*= (numeric m)
{
  word l = 0;
  while (l < n)
    item[l++] *= m;
  return *this;
}

pointset const &pointset::operator+= (point const &p1)
{
  word l = 0;
  while (l < n)
    item[l++] += p1;
  return *this;
}

pointset const &pointset::operator-= (point const &p1)
{
  word l = 0;
  while (l < n)
    item[l++] -= p1;
  return *this;
}

pointset const &pointset::operator= (pointset const &ps)
{
  delete[] item;
  n = ps.n;
  if (n) {
    point::default_dimension = ps.item[0].d;
    item = new point[n];
  } else {
    item = NULL;
  }
  epsilon = ps.epsilon;
  order = ps.order;
  word l = 0;
  while (l < n) {
    item[l] = ps.item[l];
    ++l;
  }
  if (ps.reg_idx) {
    reg_idx = new word[order];
    for (word l = 0; l < order; ++l)
      reg_idx[l] = ps.reg_idx[l];
  } else {
    reg_idx = NULL;
  }
  return *this;
}

void pointset::metric (FILE *f) const
{
  fprintf (f, "# %u x %u distance matrix\n#\n", n, n);
  for (word i = 0; i < n; ++i) {
    fprintf (f, "# ");
    for (word j = 0; j < n; ++j)
      fprintf (f, "%8.2f ", (item[i] - item[j]) . length());
    fprintf (f, "\n");
  }
}

word pointset::get (FILE *fp)
{
  word count = 0;
  SimpleList<point> pts;
  byte dim = 0;
  bool comment = false;
  bool points = false;
  int c;
  while ((c = fgetc (fp)) != EOF) {
    ++count;
    if (c == '#') {
      comment = true;
    } else if (comment) {
      if (c == '\n')
	comment = false;
    } else if (points) {
      if (c == '(') {
	ungetc (c, fp);
	point *p = new point (0);
	count += p->get (fp) - 1;
	if (dim == 0) {
	  dim = p->d;
	} else if (dim != p->d) {
	  cerr << progname << ": dimension mismatch in pointset\n";
	  exit (1);
	}
	pts.add_last (p);
      } else if (c == '}') {
	break;
      } else if (!isblank (c) && c != '\n') {
	cerr << "C: " << c << endl;
	cerr << progname << ": error in input file (pointset format)\n";
	exit (1);
      }
    } else {
      if (c == '{')
	points = true;
    }
  }

  if (pts.length () == 0) {
    cerr << progname << ": error in input file (no pointset)\n";
    exit (1);
  }

  reset (pts.length (), pts.first () -> d);

  word l = 0;
  point *p = pts.first ();
  while (p) {
    item[l++] = *p;
    p = pts.next ();
  }
  pts.purge ();

  return count;
}

word pointset::put (bool header, bool xfig, char *file) const
{
  FILE *fp;
  if (file) {
    fp = fopen (file, "w");
    if (!fp) {
      cerr << progname << ": failed to open file " << file << endl;
      exit (1);
    }
  } else {
    fp = stdout;
  }

  word count = 0;
  if (xfig) {
    count += xfig_header (fp, xfig_version);
    count += xput (fp);
  } else if (header) {
    count += fprintf (fp, "# FReg-1.0\n");
    count += put (fp);
  }

  if (file)
    fclose (fp);

  return count;
}

// Write the pointset to a file with a bit of human formating
word pointset::put(FILE *f, char *s1, char *s2) const
{
  word count = 0;
  if(s1)
    count += fprintf (f,"%s",s1);
  count += fprintf (f,"{\n");

  word l; 
  for(word k = 0; k < n; ++k) {
    count += fprintf (f, "  ");
    count += (l = item[k].put (f));
    while (l++ < 45)
      count += fprintf(f," ");
    count += fprintf (f, "# point %d\n", k);
  }

  count += fprintf (f, "}\n");
  if (s2)
    count += fprintf (f, "%s", s2);

  return count;
}

// write the pointset to a string buffer
word pointset::sput(char *f, char *s1, char *s2) const
{
  word count = 0;
  if(s1)
    count += sprintf (f, "%s", s1);
  count += sprintf (f, "{");
  for (word i = 0; i < n; ++i)
    count += item[i].sput (f);
  count += sprintf (f, "}");
  if (s2)
    count += sprintf (f, "%s", s2);
  return count;
}

// Write the pointset as xfig records to a file, dont include header info
word pointset::xput (FILE *f) const
{
  numeric min_x = cmin(0);
  numeric max_x = cmax(0);
  numeric min_y = cmin(1);
  numeric max_y = cmax(1);
  numeric min_z = cmin(2);
  
  int centre_x = xfig_wide / 4;
  int centre_y = xfig_high / 4;
  
  point centre;
  centre.set3 (centre_x, centre_y, 0.0);
  
  numeric scale_x = xfig_wide / (cwid (0) + 1.0);
  
  point corner;
  corner.set3 (min_x, min_y, min_z);
  
  point lens;
  lens.set3 (3.0 * max_x * scale_x / 4.0,
	     3.0 * max_y * scale_x / 4.0,
	     -1.0 * cwid (2) * scale_x);

  point focus;
  focus.set3 (3.0 * max_x * scale_x / 4.0,
	      3.0 * max_y * scale_x / 4.0,
	      -2.0 * cwid(2) * scale_x);
  
  pointset fred = project ((*this - corner) * scale_x, focus, lens) + centre;
  
  char str[100];
  word count = 0;
  for (word i = 0; i < n; ++i) {
    xfig_point_record xpoint;
    xfig_text_record xtext;
    xpoint.standard (fred.item[i]);
    count += xpoint.put (f);
    item[i].sput (str);
    xtext.standard (fred.item[i], str);
    count += xtext.put (f);
  } 

  for (word i = 0; i < n; ++i) {
    xfig_line_record xline;
    xline.standard (fred.item[i], fred.item[i+1]);
    count += xline.put(f);
  }

  return count;
}

// positively shift a whole pointset by the amount specified
pointset operator+ (pointset const &ps, point const &p)
{
  pointset q (ps.n, ps.n);
  word l = 0;
  while (l < ps.n) {
    q.item[l] = ps.item[l] + p;
    ++l;
  }
  return q;
}

// negatively shift a whole pointset by the amount specified
pointset operator-(pointset const &ps, point const &p)
{
  pointset q (ps.n, ps.n);
  word l = 0;
  while (l < ps.n) {
    q.item[l] = ps.item[l] - p;
    ++l;
  }
  return q;
}

pointset operator*(pointset const &ps, numeric x)
{
  pointset q (ps.n, ps.n);
  word l = 0;
  while (l < ps.n) {
    q.item[l] = ps.item[l] * x;
    ++l;
  }
  return q;
}

pointset operator/(pointset const &ps, numeric x)
{
  pointset q (ps.n, ps.n);
  word l = 0;
  while (l < ps.n) {
    q.item[l] = ps.item[l] / x;
    ++l;
  }
  return q;
}

// Project a pointset. The focus of the projection is focus, the
// closest point on the projection plane to focus is lens. Points 
// in the pointset that are too close to focus are ignored.
pointset project (pointset const &p, point const &focus, point const &lens)
{
  pointset q (p.n, p.n);
  point dirn = signum (lens - focus);
  
  if (dirn.d > 0) {

    word l = 0;
    while (l < p.n) {

      point z = dirn * (dirn % (p.item[l] - lens));
      
      point xy = p.item[l] - lens - z;
      
      numeric a = (lens - focus) . length();
      numeric b = z.length();
      
      xy = xy * (b / (a + b));
      
      q.item[l++] = z + xy + lens;
      
    }
  }
  
  return q;
}
