// points.hxx: point classes.
//
// Copyright (C) 2000 Bruce I Mills <B.I.Mills@cs.cf.ac.uk>
// Copyright (C) 2001-2002 Frank C Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.

// Geometric point class
//
// If you want to get technical about it, this is a combination of a
// number of general ideas about affine spaces. These elements are
// points in space, but also admit vector type operations such as
// multiplying a point by two (resulting in a point twice as far
// from the origin). These types of operations technically don't
// make any sense for a point intrinsically, but rather are specifically
// tied to the coordinate system. Same sort of thing goes for getting
// the length of a point and so on. However, if we take it that there
// is a prefered coordinate system, in which the transforms are going
// to be expressed, then these routines give general operations that
// can be used to construct various co-ordinate specific operations
// that are useful in a utilitarian manner for analysing the point
// sets.
//
// The point class member functions
//
// The text format of a point is a sequence of floating point numbers
// with spaces between them surrounded by parenthesese.
//
// For example ....   ( 12.3 45.6 12.9 )
//
// The space of 0 dimensions has only one point, expressed as ().
//
// Although the emphasis is on points in three space, and some of the
// routines are not generalised to different numbers of dimensions, 
// whenever it was easy to do this I have. Further, at the very least
// points in two and one dimensional space can be useful at times for
// keeping track of the results of projections, or the shape of a
// face within its carrier geometry. Points with up to 4 dimensions
// make sense when you are using `distance from a set of points'
// co-ordinates, as is done at various times in the code.

class point
{
public:

  byte d;         // the dimension of the point
  numeric *coord; // the coordinates of the point

  static byte default_dimension;

  point () {
    d = default_dimension;
    if (d > 0)
      coord = new numeric[d];
    else
      coord = NULL;
  }
  point (byte dim) {
    if (dim > 0)
      coord = new numeric[dim];
    else
      coord = NULL;
    d = dim;
  }
  point (point const &p) {
    d = p.d;
    if (d > 0) {
      coord = new numeric[d];
      byte l = 0;
      while (l < d)
	coord[l++] = p.coord[l];
    } else {
      coord = NULL;
    }
  }

  ~point () {
    if (coord)
      delete[] coord;
  }

  void set3 (numeric a, numeric b, numeric c) {
    if (d != 3) {
      d = 3;
      if (coord)
	delete[] coord;
      coord = new numeric[d];
    }
    coord[0] = a;
    coord[1] = b;
    coord[2] = c;
  }

  numeric length() const; // compute its distance from the origin

  point const &operator= (point const &p);
  point const &operator*= (numeric m);
  point const &operator+= (point const &p1);
  point const &operator-= (point const &p1);

  // Get point from a file
  word get (FILE *fp = stdin);

  // write the point to a file with start and end formating strings
  word put (FILE *f = stdout, char *s1 = NULL, char *s2 = NULL) const;
  // write the point to a string with start and end formating strings
  word sput(char *f, char *s1 = NULL, char *s2 = NULL) const;
};

point origen (byte n);
point signum (point const &p);
point operator+ (point const &p1, point const &p2);
point operator- (point const &p1, point const &p2);
point operator/ (point const &p1, numeric x);
point operator* (point const &p1, numeric x);
inline point operator* (numeric x, point const &p1) { return p1 * x; };
numeric operator% (point const &p1, point const &p2);

// The pointset class definition
//
// Like the point class, which can be a geometric point, a vector or
// a 1-form, deopending on the context, a pointset is similarly a
// multi-talented entity. Apart from being a collection of geometric 
// points it also represents simplices and n-forms in m-space.
// 
// The order variable in principle contains the dimension of the
// pointset, if it is well defined. Ie, if all the points are
// certain to have the same dimension, then this is recorded in the 
// order, however this is for those operations that operate on 
// an effecitive subclass of pointsets. A pointset might happen
// to have all the points of the same dimension even if this
// fact is not recorded in the order information. 
// 
// In principle a pointset might have points of varying dimensions, 
// I have not used this option explicitly, but there seems to be no
// reason for specifically disallowing it. If you wanted to though,
// just set the dimension when a point is put into it first, and
// then check the dimension from then on, generating an error return
// whenever it does not match.
// 
// The textual format of the poinset is { <point> <point> ... }
// 
// For example {(1.2 3.4 5.1) (4.2 3.4 5.6)}

class pointset
{
public:
  word n;                          // size of the pointset
  point *item;                     // nominal position of each point
  numeric epsilon;                 // tolerance of the point positions
  byte order;                      // the dimension of the points
  word *reg_idx;                   // registration set indices

  pointset (word max_size = 0, byte dim = 0) {
    if (max_size && dim) {
      point::default_dimension = dim;
      item = new point [max_size];
      n = max_size;
    } else {
      item = NULL;
      n = 0;
    }
    reg_idx = NULL;
  }
  pointset (char *file);
  pointset (pointset const &ps) {
    n = ps.n;
    if (n) {
      point::default_dimension = ps.item[0].d;
      item = new point [n];
    } else {
      item = NULL;
    }
    epsilon = ps.epsilon;
    order = ps.order;
    word l = 0;
    while (l < n) {
      item[l] = ps.item[l];
      ++l;
    }
    if (ps.reg_idx) {
      reg_idx = new word[order];
      for (word l = 0; l < order; ++l)
	reg_idx[l] = ps.reg_idx[l];
    } else {
      reg_idx = NULL;
    }
  }

  ~pointset () {
    if (item)
      delete[] item;
    if (reg_idx)
      delete[] reg_idx;
  }

  void reset (word max_size = 0, byte dim = 0) {
    if (item)
      delete[] item;
    if (max_size && dim) {
      point::default_dimension = dim;
      item = new point [max_size];
      n = max_size;
    } else {
      item = NULL;
      n = 0;
    }
  }

  // Determine some characteristics of the pointset
  numeric cmin (int n) const;   // minimum value of co-ordinate
  numeric cmax (int n) const;   // maximum value of co-ordinate
  numeric cwid (int n) const;   // span of co-ordinate (max - min)
  numeric character () const;   // a characteristic length
  point centroid () const;      // the centroid of the set of points
  numeric measure (word m, word k, point const &c) const;

  // Multiply pointset with constant
  pointset const &pointset::operator*= (numeric m);

  // Add point to all points in pointset
  pointset const &pointset::operator+= (point const &p1);

  // Sub point to all points in pointset
  pointset const &pointset::operator-= (point const &p1);

  pointset const &pointset::operator= (pointset const &ps);

  // Write the metric to a file
  void metric(FILE *f = stdout) const;

  // Read points from stdin (next in file should be points)
  word get (FILE *fp = stdin);

  // Write pointset to named file or stdout by default
  word put (bool header, bool xfig = false, char *file = NULL) const;

  // Write pointset to file
  word put (FILE *f = stdout, char *s1 = NULL, char *s2 = NULL) const;
  // Write pointset to string
  word sput (char *f, char *s1 = NULL, char *s2 = NULL) const;
  // write the pointset to a file, in xfig format
  word xput (FILE *f) const;
};

pointset operator+ (pointset const &ps, point const &p);
pointset operator- (pointset const &ps, point const &p);
pointset operator* (pointset const &ps, numeric x);
pointset operator/ (pointset const &ps, numeric x);
pointset project (pointset const &p, point const &focus, point const &lens);

inline pointset operator* (numeric p, pointset const &ps) {
  return ps * p;
}

inline pointset operator/ (pointset const &p1, word x) {
  return p1 / static_cast<numeric>(x);
}

inline numeric sqr (numeric x) {
  return x * x;
}
