// realisation.cc: model for symmetry detection.
//
// Copyright (C) 2000 Bruce I Mills <B.I.Mills@cs.cf.ac.uk>
// Copyright (C) 2001-2002 Frank C Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.

#include "config.h"
#include "list.hxx"
#include "xfig.hxx"
#include "points.hxx"
#include "realisation.hxx"
#include "build.hxx"

// ==================== The dualedge class ==============================

// Write a dualedge to a file
word dualedge::put (FILE *f, char *s1, char *s2) const
{
  word count = 0;
  if (s1)
    count += fprintf (f, "%s", s1);
  count += fprintf (f, "  (%d:%d %d:%d)", vertex[0], vertex[1],
		    face[0], face[1]);
  if (s2)
    count += fprintf (f, "%s", s2);
  return count;
}

// Write a dualedge to a string
word dualedge::sput (char *f, char *s1, char *s2) const
{
  return sprintf (f, "(%d:%d %d:%d)", vertex[0], vertex[1], face[0], face[1]);
}

word dualedge::get (FILE *f)
{
  word count = 0;
  bool de = false;
  bool v0 = false;
  bool v1 = false;
  bool f0 = false;
  bool f1 = false;
  char buffer[512];
  int c;
  while ((c = fgetc (f)) != EOF) {
    ++count;
    if (de) {
      if (isdigit (c)) {
	word l = 0;
	do {
	  buffer[l++] = c;
	  if (l > 511) {
	    cerr << progname << ": format error (dualedge)\n";
	    exit (1);
	  }
	  c = fgetc (f);
	  ++count;
	} while (isdigit (c));
	buffer[l] = '\0';
	ungetc (c, f);
	--count;
	if (!v0) {
	  v0 = true;
	  vertex[0] = atol (buffer);
	} else if (!v1) {
	  v1 = true;
	  vertex[1] = atol (buffer);
	} else if (!f0) {
	  f0 = true;
	  face[0] = atol (buffer);
	} else if (!f1) {
	  f1 = true;
	  face[1] = atol (buffer);
	} else {
	  cerr << progname << ": format error (dualedge)\n";
	  exit (1);
	}
      } else if (c == ')') {
	de = false;
	break;
      } else if (!isblank (c) && c != ':' & c != '\n') {
	cerr << progname << ": dualedge format error\n";
	exit (1);
      }
    } else {
      if (c == '(')
	de = true;
    }
  }
  
  if (de) {
    cerr << progname << ": dualedge format error\n";
    exit (1);
  }

  return count;
}

// ================== The dualgraph class =============================

dualgraph::dualgraph (dualgraph const &dg)
{
  n = dg.n;
  if (n) {
    item = new dualedge[n];
    word l = 0;
    while (l < n) {
      item[l] = dg.item[l];
      ++l;
    }
  } else {
    item = NULL;
  }
  def = dg.def;
}

// initialise to the complete graph of k nodes
// ie put in a link between every pair of nodes, for k nodes

void dualgraph::complete (word k)
{
  reset ((k * (k - 1)) >> 1);
  n = 0;
  for (word i = 0; i < k; ++i) {
    for (word j = i + 1; j < k; ++j) {
      item[n].vertex[0] = i;
      item[n].vertex[1] = j;
      item[n].face[0] = 0;
      item[n].face[1] = 0;
      n++;
    }
  }
  return;
}

// Fudge anyone?
//
// At the time of writing this comment, the full dualgraph 
// structure is not being used, because the face information
// is not being used. So the dualgraph is mostly being used
// as a graph, or if you prefer ... simply storing the 
// skeleton. So I happened to need to know the number of
// vertices. Thus this function just goes and counts them.
// It might be an idea to memorize this value, and maybe 
// work it out as the dual graph is built. But this routine
// does get the right answer.

int dualgraph::nodes ()
{
  int base[2 * n];
  
  word l = 0;
  while (l < 2 * n)
    base[l++] = 0;
  l = 0;
  while (l < n) {
    base[item[l].vertex[0]] = 1;
    base[item[l++].vertex[1]] = 1;
  }
  
  word t = 0;
  l = 0;
  while (l < 2 * n)
    t += base[l++];
  
  return t;
}

word dualgraph::minface() const
{
  word minl = item[0].face[0];
  word minr = item[0].face[1];
  for (word i = 0; i < n; ++i) {
    minl = (minl>item[i].face[0] ? item[i].face[0] : minl);
    minr = (minr>item[i].face[1] ? item[i].face[1] : minr);
  }

  word min = (minl>minr ? minr : minl);

  return min;
}

word dualgraph::maxface() const
{
  word maxl = item[0].face[0];
  word maxr = item[0].face[1];
  for(word i=0;i<n;i++) {
    maxl = (maxl<item[i].face[0] ? item[i].face[0] : maxl);
    maxr = (maxr<item[i].face[1] ? item[i].face[1] : maxr);
  }

  word max = (maxl<maxr ? maxr : maxl);

  return max;
}

dualgraph const &dualgraph::operator= (dualgraph const &dg)
{
  n = dg.n;
  if (item)
    delete[] item;
  if (n) {
    item = new dualedge[n];
    word l = 0;
    while (l < n) {
      item[l] = dg.item[l];
      ++l;
    }
  } else {
    item = NULL;
  }
  def = dg.def;
  return *this;
}

// Print human readable comments about the face loops.  Actually, if
// the faces are not numbered consecutively then this routine invents
// empty faces to pad out.  Could do something about this if you want
// to.

word dualgraph::note (FILE *f, char *s1, char *s2) const
{
  word count = 0;
  if (s1)
    count += fprintf (f, "%s", s1);

  // figure out the range of faces
  
  word maxl = item[0].face[0];
  word maxr = item[0].face[1];
  word minl = item[0].face[0];
  word minr = item[0].face[1];
  for (word i = 0; i < n; ++i) {
    maxl = (maxl<item[i].face[0] ? item[i].face[0] : maxl);
    maxr = (maxr<item[i].face[1] ? item[i].face[1] : maxr);
    minl = (minl>item[i].face[0] ? item[i].face[0] : minl);
    minr = (minr>item[i].face[1] ? item[i].face[1] : minr);
  }

  word max = (maxl<maxr ? maxr : maxl);
  word min = (minl>minr ? minr : minl);
  
  // now print an entry for each face in the range
  
  for (word j = min; j <= max; ++j) {
    count += fprintf (f, "# face %3d ... ", j);
    for (word i = 0; i < n; ++i) 
      if (item[i].face[0] == j || item[i].face[1] == j)
	count += fprintf (f, " %2d", i);
    count += fprintf (f, "\n");
  }
  
  if (s2)
    count += fprintf (f, "%s", s2);

  return count;
}

word dualgraph::put (FILE *f, char *s1, char *s2) const
{
  word count=0;
  if (s1)
    count += fprintf (f, "%s", s1);

  count += fprintf (f, "[\n");
  for (word i = 0; i < n; ++i) {
    word l;
    count += (l = item[i].put (f));
    while(l++ < 45)
      count += fprintf(f," ");
    fprintf(f,"# edge %d\n",i);
  }
  count += fprintf (f, "]\n");

  if (s2)
    count += fprintf (f, "%s", s2);

  return count;
}

// write the dualgraph as text, to a string

word dualgraph::sput (char *f)
{
  word count = 0;
  count += sprintf (f, "[");
  for (word i = 0; i < n; ++i)
    count += item[i].sput (f);
  count += sprintf (f, "]");
  return count;
}

// Write the dualgraph as xfig records, since a dual graph contains
// only combinatorial information, we need the realisation as well,
// this information is given as a pointset. More specificaly, the
// information in the dualgraph, at least that part being used
// at the moment, is a collection of links between vertices. But,
// each vertice is simply represented by an integer. This integer
// is then interpreted as an index into the list of points in the
// pointset, enabling the geometry of the points to be determined.

void dualgraph::xput(char *name, pointset p)
{
  FILE *f;
  f = fopen (name, "w");
  if (!f)
    return;
  xfig_header (f, xfig_version);
  xput (f, p);
  fclose (f);
}

void dualgraph::xput (FILE *f, pointset p)
{
  numeric min_x = p.cmin(0);
  numeric max_x = p.cmax(0);
  numeric min_y = p.cmin(1);
  numeric max_y = p.cmax(1);
  numeric min_z = p.cmin(2);
  
  int centre_x = xfig_wide / 4;
  int centre_y = xfig_high / 4;
  
  point centre;
  centre.set3 (centre_x, centre_y, 0);

  numeric scale_x = xfig_wide / (p.cwid(0) + 1.0);
  
  point corner;
  corner.set3 (min_x, min_y, min_z);
  
  point lens;
  lens.set3 (3.0 * max_x * scale_x / 4.0,
	     3.0 * max_y * scale_x / 4.0,
	     -1.0 * (p.cwid(2) + p.cwid(0)) * scale_x);
  
  point focus;
  focus.set3 (3.0 * max_x * scale_x / 4.0,
	      3.0 * max_y * scale_x / 4.0,
	      -2.0 * (p.cwid(2) + p.cwid(0)) * scale_x);
 
  pointset fred = project ((p - corner) * scale_x, focus, lens) + centre;
  
  // write out all the points in the pointset
  
  for (word i = 0; i < p.n; ++i) {
    xfig_point_record xpoint;
    xfig_text_record xtext;
    char str[100];
    xpoint.standard(fred.item[i]);
    xpoint.put(f);
    p.item[i].sput(str);
    xtext.standard(fred.item[i],str);
    xtext.put(f);
  }

  // connect the dots with the implicit line segments
  
  for (word i = 0; i < n; ++i) {
    xfig_line_record xline;
    xline.standard(fred.item[item[i].vertex[0]],fred.item[item[i].vertex[1]]);
    xline.put(f);
  }

  return;
}

word dualgraph::get(FILE *f)
{
  def = false;

  if (!f) {
    reset (0);
    return 0;
  }

  SimpleList<dualedge> edges;

  word count = 0;
  bool dg = false;
  bool comm = false;
  int c;
  while ((c = fgetc (f)) != EOF) {
    ++count;
    if (comm) {
      if (c == '\n')
	comm = false;
    } else if (dg) {
      if (c == '(') {
	ungetc (c, f);
	--count;
	dualedge *de = new dualedge ();
	count += de->get (f);
	edges.add_last (de);
      } else if (c == ']') {
	dg = false;
	break;
      } else if (c == '#') {
	comm = true;
      } else if (!isblank (c) && c != '\n') {
	cerr << progname << ": dualgraph format error\n";
	exit (1);
      }
    } else {
      if (c == '[') {
	def = true;
	dg = true;
      } else if (c == '#') {
	comm = true;
      }
    }
  }

  if (dg) {
    cerr << progname << ": dualgraph format incomplete error\n";
    exit (1);
  }

  reset (edges.length ());

  word l = 0;
  dualedge *de = edges.first ();
  while (de) {
    item[l++] = *de;
    de = edges.next ();
  }
  edges.purge ();

  return count;
}

// ====================== the realisation class ==========================

// Read a realisation from a file

realisation::realisation (char *file)
{
  ftype = NULL;
  max_face = 0;

  FILE *fp;
  if (file) {
    fp = fopen (file, "r");
    if (!fp) {
      cerr << progname << ": cannot read file " << file << endl;
      exit (1);
    }
  } else {
    fp = stdin;
  }

  char buffer[512];
  fgets (buffer, 511, fp);
  if (strncmp (buffer, "# FReg-1.0", 10) == 0) {
    get (fp);
  } else if (strncmp (buffer, "#FIG", 4) == 0) {
    xget (fp);
  } else {
    cerr << progname << ": unkown file format\n";
    fclose (fp);
    exit (1);
  }

  fclose (fp);
}

class ftpair {
public:
  word fn, ft;
};

word realisation::get (FILE *f)
{
  if (!f)
    return 0;

  word count = geometry.get (f);
  count += scheme.get (f);

  // Read type information, if present...

  ftype = NULL;

  SimpleList<ftpair> pairs;
  
  char buffer[512];
  bool comm = false;
  bool ty = false;
  ftpair *ft = NULL;
  max_face =0;
  int c;
  while ((c = fgetc (f)) != EOF) {
    ++count;
    if (comm) {
      if (c == '\n')
	comm = false;
    } else if (ty) {
      if (isdigit (c)) {
	word l = 0;
	do {
	  buffer[l++] = c;
	  if (l > 511) {
	    cerr << progname << ": format error (face types)\n";
	    exit (1);
	  }
	  c = fgetc (f);
	  ++count;
	} while (isdigit (c));
	buffer[l] = '\0';
	ungetc (c, f);
	--count;
	if (ft) {
	  ft->ft = atol (buffer);
	  pairs.add_last (ft);
	  ft = NULL;
	} else {
	  ft = new ftpair;
	  ft->fn = atol (buffer);
	  if (ft->fn > max_face)
	    max_face = ft->fn;
	}
      } else if (c == '>') {
	ty = false;
	break;
      } else if (c == '#') {
	comm = true;
      } else if (!isblank (c) && c != ',' && c != ';' && c != '\n' &&
		 c != '~') {
	cerr << progname << ": format error (face types)\n";
	exit (1);
      }
    } else {
      if (c == '<')
	ty = true;
      else if (c == '#')
	comm = true;
    }
  }

  if (ft || ty) {
    cerr << progname << ": face type information incomplete\n";
    exit (1);
  }

  if (pairs.length () > 0) {
    ++max_face;
    if (ftype)
      delete[] ftype;
    ftype = new word[max_face];
    word l = 0;
    while (l < max_face)
      ftype[l++] = 0;
    ft = pairs.first ();
    while (ft) {
      ftype[ft->fn] = ft->ft;
      ft = pairs.next ();
    }
  }

  return count;
}

// look only for circles and lines.
// 
// Each circle is a point, and each line is a link.
//
// The centre of the circle directly gives the location of the
// point. However it is invalid for the centre of one circle to
// be inside the radius of another. The end point of a line is
// said to be the same as a point if it is inside the radius of
// the circle. 
//
// xfig types ....
//
// 1 = circle
// 2 = line

class line {
public:
  line () : p1 (2), p2 (2) {
  }
  point p1, p2;
};

word realisation::xget (FILE *f)
{
  if (!f)
    return 0;

  SimpleList<point> pts;
  SimpleList<line> lns;

  int n;
  while ((n = fgetc (f)) != EOF) {

    if (isblank (fgetc (f))) {

      if (n == '1') {

	xfig_point_record xpoint;
	xpoint.get(f);
	
	point *p = new point (3);
	p->coord[0] = xpoint.field12;
	p->coord[1] = xpoint.field13;
	p->coord[2] = xpoint.field14;
	
	pts.add_last (p);
	
      } else  if (n == '2') {

	xfig_line_record xline;
	xline.get(f);
	
	line *l =new line ();
	l->p1.coord[0] = xline.field16;
	l->p1.coord[1] = xline.field17;
	l->p2.coord[0] = xline.field18;
	l->p2.coord[1] = xline.field19;

	lns.add_last (l);
	
      }
    }
    while (fgetc (f) != '\n' && !feof (f));

  }

  // Setup geometry
  numeric radius[pts.length ()];
  geometry.reset (pts.length (), 2);
  word k = 0;
  point *p = pts.first ();
  while (p) {
    geometry.item[k].coord[0] = p->coord[0];
    geometry.item[k].coord[1] = p->coord[1];
    radius[k++] = p->coord[2];
    p = pts.next ();
  }
  pts.purge ();

  // Setup scheme, i.e. sort out which point is which
  k = 0;
  scheme.reset (lns.length ());
  scheme.def = (lns.length () > 0);
  line *l = lns.first ();
  while (l) {

    for (word j = 0; j < geometry.n; ++j) {
      if ((l->p1 - geometry.item[j]) . length () < radius[j])
	scheme.item[k].vertex[0] = j;
      if ((l->p2 - geometry.item[j]) . length () < radius[j])
	scheme.item[k].vertex[1] = j;
    }

    // No faces
    scheme.item[k].face[0] = 0;
    scheme.item[k].face[1] = 0;

    ++k;
    l = lns.next ();
  }
  lns.purge ();

  // No types
  if (ftype)
    delete[] ftype;
  max_face = 0;

  return 1;
}

// Write a realisation to a file
word realisation::put (FILE *f, char *s1, char *s2) const
{
  word count = fprintf (f, "# FReg-1.0\n\n");

  if (s1)
    count += fprintf (f, "%s", s1);
  
  count += geometry.put (f);

  if (scheme.def) {
    count += fprintf (f, "\n# List of edges on each face\n#\n");
    count += scheme.note (f);
    count += fprintf (f, "\n");
    count += scheme.put (f);
  }

  bool ft = false;
  word l = 0;
  while (l < max_face) {
    if (ftype[l++]) {
      ft = true;
      break;
    }
  }

  if (ft) {
    count += fprintf(f,"\n# List of non type-0 faces\n<\n");
    l = 0;
    while (l < max_face) {
      if (ftype[l]) {
	word len = fprintf (f, "  %d ~ %d", l, ftype[l]);
	while (len++ < 45)
	  fprintf (f, " ");
	count += len + fprintf (f, "# face %d\n", l);
      }
      ++l;
    }
    fprintf (f, ">\n");
  }

  if(s2)
    count += fprintf (f, "%s", s2);
  
  return count;
}
