// Realisation.hxx: model for symmetry detection.
//
// Copyright (C) 2000 Bruce I Mills <B.I.Mills@cs.cf.ac.uk>
// Copyright (C) 2001-2002 Frank C Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.

// Precis of the types defined in this file ....
//
// class point ............... a single geometric point
// class pointset ............ a collection of points
// class dualedge ............ an edge in a dual graph
// class dualgraph ........... two graphs with the same edge set
// class realisation ......... scheme and geometry of a B-rep
// class pseudosym ........... a symmetry of a dualgraph
// class injection ........... a one to one integer function
// class heap ................ a priority queue implimentation
// class partition ........... a partition of a pointset
//
// Each of these types is implemented directly with arrays. Sometimes
// the use of the array indices uses the same approach as various
// linked list algorithms. Remember that, logically speaking, an index
// into an array IS a pointer. It would probably not be advisable to
// adjust this stuff to use internal linked lists. Firstly because it
// would mean that you would rely heavily on the compilers version of
// memory handling, and in my experience there are typically faults in
// these, which lead to mysterious crashes. If you want to write your
// own memory handler ... go ahead. Secondly, it would make each
// routine somewhat less self contained, harder to check exactly what
// the individual operations are doing. Thirdly, at least some of them
// will not get the same performance as a linked list. For example,
// the injection type trades off speed and memory to get constant time
// operations, that would be much slower if implemented for indefinite
// extension using dynamically allocated memory.
//

// ============= Beginning of the type definitions ===================

// Each of the types has a similar structure. Some basic once off
// data, and an array or two of something. So, for example, a point
// has a dimension and an array of co-ordinates, and a pointset has a
// size, and an array of points. A lot of the routines available are
// the same, or very close. So, for example, put, get, sput, all exist
// for each data type, and come with a fairly standard collection of
// argument lists depending on where you want to put the type, or
// where you are reading it from. All the types have human readible
// text forms, that are fully catered for, so you can read and write
// them to human editable files. Further, there are a variety of infix
// operators defined, so for example you can multiple two points by
// each other to get a inner product, or multiply a pointset by a
// float to change its scale. Check these out, and use them as much as
// possible. It makes the code easier to followe, and more
// robust. Also, as noted below, comments are introduced by `#'.

// ============= Basic text routines to help with parsing ============

// I have tried to conform to the idea that each type has a written
// form, in this way they can be writen two and read from text files.
// The textual form of the types is reasonably straight forward, and
// can be easily generated and edited by a human being. The format is
// not rigid at all, which makes editing reasonably easy, and allows
// the user to decide how to lay things out on the page for their own
// use, a `#' symbol is used to introduce a comment that lasts until
// the end of the input line. There are two forms in which each data
// type can be output. One is cursive, and the other is an expanded 
// form which includes simple comments. For example, the expanded 
// pointset format labels each point with an integer indicating its
// position in the list. I have found this useful at times, keeping
// track of how many points in the set etc. It is also possible to
// output a lot of this information to an xfig file, and input it
// back again. However not all information is retained.
//
// The actual text formats for the classes are explained in the
// sections defining those classes.
//
// The basic naming is ...
//
// put .... to do a quick write to a file
// sput ... do a write to string
// hput ... human friendly write to a file
// get .... read from a file
//
// The routine `put' can take arguments such as a file pointer 
// or a char pointer (to the name of the file). One reason for
// writing to a string being different (apart from having to
// use sprintf, rather than fprintf) is that a char pointer 
// as first argument to put means that it is the name of a file,
// while a char pointer to sput is the place that the information
// will be stored. This is not naturally distinguished by the
// argument type list, and thus the use of distinct names.


// ==================== The dualedge class ==============================

// The dual edge exists in order to be able to construct the dualgraph
// type. The dual graph is an implimentation of a combinatorial scheme
// restricted to three dimensions. In this case the information that
// we need to store is, which vertices are the end points of which
// edges, and which edges are on the boundary of which faces. Since in
// a valid model an edge joins two nodes (maybe the same one to
// itself), and an edge also joins (glues, so to speak) two faces
// together, the information can be stored as two graphs. The edges
// and vertices form the vertex graph (also called the skeleton of the
// boundary rep) and the edges and faces form the face graph. The only
// problem is that without knowing the topology of the geometric
// surface, and how the embedding occurs, there is more than one
// possible interpretation of all this. So we need to remember that
// the edges in one graph are actually the same edges as in the
// other. This is handled here by making each edge store a pair of
// vertices, and a pair of faces.
//
// The input/output of dualedges in text form is rather cursive.  You
// might like to add the missing routines. But I just needed the
// basics to make it work, so that's all that is here at the moment.
//
// A dual edge is basically a pair of pairs of integers, its format is
// ( <int>:<int> <int>:<int>)
//
// For example (1:2 2:4) is the dualedge that connects vertex 1 to
// vertex 2, and face 2 to face 4. Ie, the first pair is the vertex
// pair, and the second pair is the face pair.

class dualedge
{
public:
  word vertex[2];  // connects vertices with these numbers
  word face[2];    // connects faces with these numbers

  dualedge const &operator= (dualedge const &de) {
    vertex[0] = de.vertex[0];
    vertex[1] = de.vertex[1];
    face[0] = de.face[0];
    face[1] = de.face[1];
    return *this;
  }

  // basic output to file routines
  word put (FILE *f = stdout, char *s1 = NULL, char *s2 = NULL) const;
  
  word sput (char *f, char *s1 = NULL, char *s2 = NULL) const;
  
  word get (FILE *f = stdin);       // read it via this file pointer
};


// This class should be understood in conjunction with the dualedge 
// class, which really only exists to make the dual graph class work.
// The dual graph impliments a specialisation of a combinatorial 
// scheme to three dimensions. For more information about what this
// means, see the comments for the dualedge class.
//
// Basically a dualgraph is a list of dual edges.
//
// At the time of writing this comment, the face information from 
// the boundary representation is not being used. It really should
// be. This code is not structured simply for pointsets, it is really
// intended to handle faces as well. So, if you are implementing 
// faces, remember that this is where the information fits.
//
// In textual format the dualgraph is a bit like a pointset, except
// that it is a list of dualedges rather than points. To distinguish
// the two, the dualgraph uses square brackets as delimiters.
// The format is [ <dualedge> <dualedge> .... ], and empty dualgraph,
// like an empty pointset, is a valid construction.
//
// Example:  [(1:2 2:3) (12:23 1:4)]

class dualgraph
{
public:
  word n;
  dualedge *item;
  bool def;

  dualgraph (word s = 0) {
    def = false;
    n = s;
    if (n)
      item = new dualedge[n];
    else
      item = NULL;
  }
  dualgraph (dualgraph const &dg);

  ~dualgraph () {
    if (item)
      delete item;
  }

  void reset (word s) {
    if (item)
      delete item;
    n = s;
    if (n) 
      item = new dualedge[n];
    else
      item = NULL;
  }

  void complete (word k); // initialise to the complete graph of k nodes

  int nodes(); // return how many (vertex) nodes involved

  dualgraph const &operator= (dualgraph const &dg);
  
  word minface() const;
  word maxface() const;

  word note(FILE *f = stdout, char *s1 = NULL, char *s2 = NULL) const;

  word put (FILE *f = stdout, char *s1 = NULL, char *s2 = NULL) const;
  word sput(char *f);

  void xput (char *name, pointset p);
  void xput (FILE *f, pointset p);
  
  word get(FILE *f);
};


// ====================== the realisation class ==========================

// This class represents a realisation of a combinatorial scheme, 
// restricted to the three dimensional case. The basic combinatorial
// information is the vertex graph and the face graph of the boundary
// representation. This information indicates which two vertices and
// edge connects, and which two faces an edge lies between. A full
// realisation of this information would mean that the location of
// each vertex, and the curved shape of each edge, and the surface
// realising each face, would have to be given. However, at the moment
// only the geometry of the points is actually stored. The geometry
// of the edges is implicitly a straight line between the two end
// points, and faces are not handled. Basically what is needed is
// a curve and surface type, and these can then be stored in the
// representation. For some more information see the comments
// about the dualgraph structure that stores the combinatorial
// information.
//
// The input/output of realisations is about as cursive as it gets.
//
// A realisation is basically an ordered pair of a dualgraph and
// a pointset. The indices in the dualedges are for indexing into
// the list of points in the pointset. At the time of writing this
// comment there is no solid realisation text format. The one 
// actually used is to just include the pointset, and then the
// dualgraph. Ie, the format is <pointset> <dualgraph>. Technically
// there is nothing wrong with this, except that it does not 
// uniquely distinguish which type of data is intended when you
// open a file and read it. However, so far this has been implicit
// from the context and pragmatics. This realy should be fixed.
// We need a symbol for introducing a realisation. One way to
// do this would be to simply write the word ``realisation''
// in front of the ordered pair.

class realisation
{
public:
  pointset geometry;
  dualgraph scheme;
  word *ftype;
  word max_face;

  realisation () {
    ftype = NULL;
    max_face = 0;
  }
  realisation (char *file);

  ~realisation () {
    if (ftype)
      delete[] ftype;
  }

  word face_type (word fn) const {
    return (fn < max_face) ? ftype[fn] : 0;
  }

  word get(FILE *f = stdin);
  word xget(FILE *f = stdin);

  word put(FILE *f = stdout, char *s1 = NULL, char *s2 = NULL) const;
};


void find_factored_regularities (FILE *f, realisation *r, bool rectify,
				 bool first_only);
