// rect.cc: rectification
//
// Copyright (C) 2000 Bruce I Mills <B.I.Mills@cs.cf.ac.uk>
// Copyright (C) 2001-2002 Frank C Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.

#include "config.h"
#include "xfig.hxx"
#include "points.hxx"
#include "realisation.hxx"
#include "cluster.hxx"
#include "maps.hxx"
#include "build.hxx"

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

// This software was developed for a project which included a directive to
// determine a method for finding partial approximate symmetries of a solid
// geometric shape. This is a core directive. But, the information actually
// provided to the software is a boundary representation model of the shape.
// Such a model suffers from the deficiency that vertices and edges might
// be an artifact of the representation rather than a feature of the solid.
// For example, a cube might be represented as having 5 square faces and
// two triangular faces (due to one of the natural square faces being made
// up from two triangular model faces). If a model is not a natural 
// representation of the solid, then it is possible that some symmetries of 
// the solid are not induced from any symmetry of the model. In the example
// of the cube given above the symmetries of the solid are induced from 
// symmetries of the vertex set. But, this will not typically be the case. 
// To recognise how to cut and glue the model faces to make them natural, 
// or at least to make sure that they reflect the symmetry of the model, 
// is in general rather difficult. Due partly to time constraints, but also
// to logical structure, I decided that it is not the place of this
// software to fix the model. Rather it is typically, but not entirely,
// assumed that the faces of the model are the natural faces of the solid.
// This software is aimed at finding partial approximate symmetry of the
// given boundary representation. Any such symmetry will be a symmetry of
// the actual geometric solid. It is just that sometimes some of the 
// symmetries of the object will be overlooked.

// Another core directive of the project is that the faces of the boundary
// representation have an underlying geometry which is a plane, cylinder,
// sphere, cone or torus. Each edge is determined as the intersection of
// the exactly two faces that contain the edge. Each vertex is determined
// as the approximate intersection of all the faces that contain the vertex.

// To this I add some extra conditioning requirements that the data should
// satisfy in order to be processed correctly by the algorithms. Each face
// needs to be at least triangular. This assumption is implicit in the code
// which determines the face permutation induced by a vertex permutation,
// when checking for the consistency of the vertex permuation with the rest
// of the combinatorial information in the model. This can be established in
// a pre-processing stage in which suitable edges are divided by an extra
// vertex placed at their mid point by arc length. By, doing this in a 
// regular manner the symmetries of the model are retained.  Each edge 
// must border at most two faces. This is already implied by the stronger 
// condition from the project directive, but I add it here, since it is an 
// implicit assumption for the correct running of some of the code. Each 
// face should be well represented by its vertices. This is not an exact 
// criteria, but in rough terms means that no point on the face should be 
// too distant from the convex hull of its vertex set. For example, any 
// polygonal face is well represented by its vertex set.

// The requirement for the face to be well represented by its vertices is
// specificaly for factoring the model to a level of tolerance. For example,
// a long thin polygonal face can be recognised as a candidate for being 
// changed to an edge, when the vertices form two groups, one at each end 
// of the face. But in the case of a roughly circular face, with three 
// vertices near each other on the edge, the vertices may form one group, 
// suggesting that the face should be replaced by a single point. Much more 
// complex processing is required to detect this sort of thing. It is 
// envisaged, but not implemented, that this can be set up in another
// pre-processing stage. For example, for any face that is not well represented
// by its vertex set, the centroid of the face can be included as a vertex
// of the face. Because of the dualgraph structure it is required that this
// vertex be included as an end point of an edge. To retain the symmetry of
// the boundary representation, and extra edge must be included between 
// each old vertex and the centroid. The geometry of the edge is indeterminate,
// and can be handled formaly in the code in this mode.

// A symmetry of the boundary model is an isometry that maps all the elements
// of the model onto compatible elements, preserving the information about 
// which element borders which other element. This software deals with this
// implicitly by storing all the information in terms of permutations of the
// vertices, edges and faces, which preserve certain geometric information.
// In particular, we begin by finding the permutations of the vertices that
// preserve the distance between all the vertices. Then we find which of those
// induce permutations of the edges and faces. The resulting permutations 
// form a collection that includes all the approximate symmetries of the
// boundary representation. Obviously, if we want an actual isometry we
// can obtain this by finding, for example, the one that has minimum 
// error in mapping the vertices in the required manner. For a natural model 
// of a solid bounded polyhedron, this produces the symmetries of the object. 
// Mapping the end points of an edge, maps the edge, and similarly mapping 
// all the edges of a face (complete with conectivity) must map the face. 
// Further a natural model of a polyhedron will have faces that each have
// at least three edges anyway. And we have seen that their faces are well
// represented by their vertices in the sense indicated above.

// However, when further types of surfaces are involved, we also have to
// check that the individual faces match. We can narrow down the permutations 
// again by checking the type of each face, if they don't match then the 
// permutation can be excluded. This leaves the problem of determining 
// matches between two surfaces of the same type. 

// By adding the centroid of each non planar face as a point in that face
// (it does not matter if the centroid is not actually on the face), and
// putting in edges that connect the centroid to each vertex of that face,
// we can generate only those combinatorial symmetries that also preserve
// the centroid of the face. It is intended that the geometric information
// about the location of the vertices, and the underlying geometry of the
// faces is of central importance. The actual geometry of the edges is not
// being considered here, so the indeterminate shape of these new edges is
// not an issue. If some matching of the shape of the edges was to be 
// undertaken, then it is a simple matter for these new edges to be considered
// as short line segements joining the end points. The fact that technically
// these will not actually lie in the face they are part of does not affect
// the correct operation of the algorithm. The net effect of these new edges
// and vertices is to convert the boundary representation into a natural
// model of a polyhedron, but with typed faces and edges. The polyhedron
// has at least the symmetries of the original model. A final post processing
// step can be done to check the mapping of the underlying geometry in order
// to determine if the symmetry refers to the original model.

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

// Here I have some code related to rectification

// The basic story is this: We have an ACIS model which is not ideal, and we
// wish to adjust it to make it better. The software here can take the 
// information about the positions of the vertices, and about which edge
// lies between which vertices, and which faces. Given this information the
// software, one way or another, produces a group of permutations of the
// vertices that preserves the distances between vertices, and all the
// structural information. It it now required to find out how to slightly
// adjust the location of the vertices so that the symmetry will be exact.

// Thus, in practice, you can extract the vertex locations, and the 
// structural information from the ACIS model, find the new locations of
// the vertices, and feed this back into the ACIS model to adjust the
// faces to get the vertices in the right place. At the time of writing
// Frank has indicated that incorporating this information in his 
// constraint solving system is straight forward. This seems to be a
// nice way to do it, since the extent to which the symmetry is rectified
// can be considered as an adjustable parameter which can be selected for
// whatever criteria to do with the constraint solving is required.

// The software in this section is aimed simply at finding the new 
// positions for the vertices given the permutation with respect to
// which they must be rectified. The result is not unique. Rather it
// is only defined up to rotation, reflection and scaling. The point
// being that it is *shape* information, and so the orientation, sense,
// and size has no particular significance. If it is desired to rescale
// afterwards (for example to keep the volume of the convex hull the 
// same) then this can be done, but it is independent of the intention
// of the software in this section.

// This routine adjusts the tetrahedron so that the appropriate
// sides are given the distances as indicated. In principle this
// is actually just the completely new construction of a tetrahedron
// with sides of specified lengths. However, it is done here by 
// fixing the points, one at a time, and the result will be close-ish
// to the original tetrahedron, if the original tetrahedron was 
// close-ish to the correct specification.

void fixup_tetrahedron (pointset *tetra,
			double dist01,
			double dist02,
			double dist03,
			double dist12,
			double dist13,
			double dist23)
{
  pointset ps (3, tetra->item[0].d);
  point p (tetra->item[0].d);
  pointset r (2, 3);
  double d0,d1;

  // fixup point 0

  ps.n = 3;
  ps.item[0] = tetra->item[1]-tetra->item[0];
  ps.item[1] = tetra->item[2]-tetra->item[0];
  ps.item[2] = tetra->item[3]-tetra->item[0];

  p.coord[0] = dist01;
  p.coord[1] = dist02;
  p.coord[2] = dist03;

  locate(&ps,&p,&r);

  d0 = r.item[0].length();
  d1 = r.item[1].length();


  tetra->item[0] = tetra->item[0] + r.item[d0<d1?0:1];
  //printf("With pt 0 fixed\n");tetra->metric();

  // fixup point 1

  ps.n = 3;
  ps.item[0] = tetra->item[0]-tetra->item[1];
  ps.item[1] = tetra->item[2]-tetra->item[1];
  ps.item[2] = tetra->item[3]-tetra->item[1];

  p.coord[0] = dist01;
  p.coord[1] = dist12;
  p.coord[2] = dist13;

  locate(&ps,&p,&r);

  d0 = r.item[0].length();
  d1 = r.item[1].length();

  tetra->item[1] = tetra->item[1] + r.item[d0<d1?0:1];
  //printf("With pt 1 fixed\n");tetra->metric();

  // fixup point 2

  ps.n = 3;
  ps.item[0] = tetra->item[0]-tetra->item[2];
  ps.item[1] = tetra->item[1]-tetra->item[2];
  ps.item[2] = tetra->item[3]-tetra->item[2];

  p.coord[0] = dist02;
  p.coord[1] = dist12;
  p.coord[2] = dist23;

  locate(&ps,&p,&r);

  d0 = r.item[0].length();
  d1 = r.item[1].length();

  tetra->item[2] = tetra->item[2] + r.item[d0<d1?0:1];
  //printf("With pt 2 fixed\n");tetra->metric();

  // fixup point 3

  ps.n = 3;
  ps.item[0] = tetra->item[0]-tetra->item[3];
  ps.item[1] = tetra->item[1]-tetra->item[3];
  ps.item[2] = tetra->item[2]-tetra->item[3];

  p.coord[0] = dist03;
  p.coord[1] = dist13;
  p.coord[2] = dist23;

  locate(&ps,&p,&r);

  d0 = r.item[0].length();
  d1 = r.item[1].length();

  tetra->item[3] = tetra->item[3] + r.item[d0<d1?0:1];

  //printf("With pt 3 fixed\n");tetra->metric();

  // and that's it ... all 6 sides should now conform to spec

}

// Rectify a pointset.
//
// The shape of the pointset is implicit in the distances between each 
// pair of vertices. In order to rectify the pointset we, in principle, 
// set each distance to the average length of all the edges that are 
// equivalent to it. However, what we need as output is an exemplar of
// the exact symmetry. So, instead we require a registration set to be
// indicated. We can then consider the collection of tetrahedra to 
// which it is mapped. Since at least localy any edge of a tetrahedron
// can be adjusted independently, we can then setup a tetrahedron to
// be the average tetrahedron. Now, for a particular point, we find
// the distance between this point and the tetrahedron, and the
// corresponding distances involved when the tetrahedron and the point
// are transformed according to the permutations. The averages of the
// corresponding distances gets us four average distances, that are 
// then used to locate the point in space with respect to the 
// rectified tetrahedron.

// The tetrahedron has to be part of the point set so we can 
// work out what the permutation is doing with it. The integer
// arguments are the indices of the tetrhedron points.


void pseudosym::rectify (pointset *ps, pointset *newps, point const &c)
{
  // first we have to brute force rectify the tetrahedron itself

  // find the average distances over all equivalent tetrahedrons

  numeric dist01 = 0.0;
  numeric dist02 = 0.0;
  numeric dist03 = 0.0;
  numeric dist12 = 0.0;
  numeric dist13 = 0.0;
  numeric dist23 = 0.0;

  for (word i = 0; i < n; ++i) {
    dist01 += (c - ps->item[item[i]->forward[ps->reg_idx[0]]]) . length();
    dist02 += (c - ps->item[item[i]->forward[ps->reg_idx[1]]]) . length();
    dist03 += (c - ps->item[item[i]->forward[ps->reg_idx[2]]]) . length();
    dist12 += (ps->item[item[i]->forward[ps->reg_idx[0]]] 
	       - ps->item[item[i]->forward[ps->reg_idx[1]]]) . length();
    dist13 += (ps->item[item[i]->forward[ps->reg_idx[1]]] 
	       - ps->item[item[i]->forward[ps->reg_idx[2]]]) . length();
    dist23 += (ps->item[item[i]->forward[ps->reg_idx[1]]] 
	       - ps->item[item[i]->forward[ps->reg_idx[2]]]) . length();
  }

  dist01 /= static_cast<numeric> (n);
  dist02 /= static_cast<numeric> (n);
  dist03 /= static_cast<numeric> (n);
  dist12 /= static_cast<numeric> (n);
  dist13 /= static_cast<numeric> (n);
  dist23 /= static_cast<numeric> (n);

  // we now have the average of the 6 inter-vertex distances 

  // printf("dist01 = %f\n",dist01);
  // printf("dist02 = %f\n",dist02);
  // printf("dist03 = %f\n",dist03);
  // printf("dist12 = %f\n",dist12);
  // printf("dist13 = %f\n",dist13);
  // printf("dist23 = %f\n",dist23);
  
  // next step ... construct a tetrahedron that matches these distances
  pointset tetra (4, ps->item[0].d);
  tetra.item[0] = c;
  tetra.item[1] = ps->item[ps->reg_idx[0]];
  tetra.item[2] = ps->item[ps->reg_idx[1]];
  tetra.item[3] = ps->item[ps->reg_idx[2]];
  // tetra.put ();
  // tetra.metric ();

  fixup_tetrahedron (&tetra, dist01, dist02, dist03, dist12, dist13, dist23);
  // tetra.put ();
  // tetra.metric ();

  // ok, now fixup every point with respect to this tetrahedron
  
  for (word j = 0; j < ps->n; ++j) {  // loop through all the points

    numeric d0 = 0.0;
    numeric d1 = 0.0;
    numeric d2 = 0.0;
    numeric d3 = 0.0;
    
    for (word i = 0; i < n; ++i) {  // loop through all the symmetries
      word g = item[i]->forward[j];
      d0 += (ps->item[g] - c).length();
      d1 += (ps->item[g] - ps->item[item[i]->forward[ps->reg_idx[0]]]) .
	length();
      d2 += (ps->item[g] - ps->item[item[i]->forward[ps->reg_idx[1]]]) .
	length();
      d3 += (ps->item[g] - ps->item[item[i]->forward[ps->reg_idx[1]]]) .
	length();
    }

    d0 /= static_cast<numeric> (n);
    d1 /= static_cast<numeric> (n);
    d2 /= static_cast<numeric> (n);
    d3 /= static_cast<numeric> (n);
    
    // we now know the average distances of this particular 
    // point from the corners of the tetrahedron
    
    pointset temp (3, ps->item[0].d);
    pointset r0 (2, 3);
    pointset r1 (2, 3);
    pointset r2 (2, 3);
    pointset r3 (2, 3);
    point f (3);
    
    temp.n = 3;

    //printf("point %d \n",j);

    //printf("frame coords = (%f %f %f)\n",d1,d2,d3);
    temp.item[0] = tetra.item[1] - tetra.item[0];
    temp.item[1] = tetra.item[2] - tetra.item[0];
    temp.item[2] = tetra.item[3] - tetra.item[0];
    f.coord[0] = d1;
    f.coord[1] = d2;
    f.coord[2] = d3;
    locate (&temp, &f, &r0);
    r0.item[0] = r0.item[0] + tetra.item[0];
    r0.item[1] = r0.item[1] + tetra.item[0];
    //r0.put ();
    
    //printf("frame coords = (%f %f %f)\n",d0,d2,d3);
    temp.item[0] = tetra.item[0] - tetra.item[1];
    temp.item[1] = tetra.item[2] - tetra.item[1];
    temp.item[2] = tetra.item[3] - tetra.item[1];
    f.coord[0] = d0;
    f.coord[1] = d2;
    f.coord[2] = d3;
    locate (&temp, &f, &r1);
    r1.item[0] = r1.item[0] + tetra.item[1];
    r1.item[1] = r1.item[1] + tetra.item[1];
    //r1.put();
    
    //printf("frame coords = (%f %f %f)\n",d0,d1,d3);
    temp.item[0] = tetra.item[0] - tetra.item[2];
    temp.item[1] = tetra.item[1] - tetra.item[2];
    temp.item[2] = tetra.item[3] - tetra.item[2];
    f.coord[0] = d0;
    f.coord[1] = d1;
    f.coord[2] = d3;
    locate (&temp, &f, &r2);
    r2.item[0] = r2.item[0] + tetra.item[2];
    r2.item[1] = r2.item[1] + tetra.item[2];
    //r2.put();
    
    //printf("frame coords = (%f %f %f)\n",d0,d1,d2);
    temp.item[0] = tetra.item[0]-tetra.item[3];
    temp.item[1] = tetra.item[1]-tetra.item[3];
    temp.item[2] = tetra.item[2]-tetra.item[3];
    f.coord[0] = d0;
    f.coord[1] = d1;
    f.coord[2] = d2;
    locate (&temp, &f, &r3);
    r3.item[0] = r3.item[0] + tetra.item[3];
    r3.item[1] = r3.item[1] + tetra.item[3];
    //r3.put();
    
    // We now have the two possible locations of the point
    // with respect to each of the four faces of the tetrahedron 
    
    // The right answers will be close to each other so find them
    
    numeric t00 = (r0.item[0] - r1.item[0]) . length();
    numeric t01 = (r0.item[0] - r1.item[1]) . length();
    numeric t10 = (r0.item[1] - r1.item[0]) . length();
    numeric t11 = (r0.item[1] - r1.item[1]) . length();

    word w0 = 0;
    word w1 = 0;
    word w2 = 0;
    word w3 = 0;

    if (t00 <= t00 && t00 <= t01 && t00 <= t10 && t00 <= t11) {
      w0=0;
      w1=0;
    }
    if (t01 <= t00 && t01 <= t01 && t01 <= t10 && t01 <= t11) {
      w0=0;
      w1=1;
    }
    if (t10 <= t00 && t10 <= t01 && t10 <= t10 && t10 <= t11) {
      w0=1;
      w1=0;
    }
    if (t11 <= t00 && t11 <= t01 && t11 <= t10 && t11 <= t11) {
      w0=1;
      w1=1;
    }
    
    t00 = (r2.item[0] - r3.item[0]).length();
    t01 = (r2.item[0] - r3.item[1]).length();
    t10 = (r2.item[1] - r3.item[0]).length();
    t11 = (r2.item[1] - r3.item[1]).length();
    
    if (t00 <= t00 && t00 <= t01 && t00 <= t10 && t00 <= t11) {
      w2=0;
      w3=0;
    }
    if (t01 <= t00 && t01 <= t01 && t01 <= t10 && t01 <= t11) {
      w2=0;
      w3=1;
    }
    if (t10 <= t00 && t10 <= t01 && t10 <= t10 && t10 <= t11) {
      w2=1;
      w3=0;
    }
    if (t11 <= t00 && t11 <= t01 && t11 <= t10 && t11 <= t11) {
      w2=1;
      w3=1;
    }
    
    // now average the four points to get the desired answer

    newps->item[j] = (r0.item[w0] + r1.item[w1] + 
		      r2.item[w2] + r3.item[w3]) / 4.0;

  }

  newps->n = ps->n;

  // fudge for testing
  
  //*newps *= 2.0 / (newps->item[0] - newps->item[1]) . length();

  return;
}
