// symmetry.cc: symmetry detection
//
// Copyright (C) 2000 Bruce I Mills <B.I.Mills@cs.cf.ac.uk>
// Copyright (C) 2001-2002 Frank C Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.

#include "config.h"
#include "xfig.hxx"
#include "points.hxx"
#include "realisation.hxx"
#include "cluster.hxx"
#include "maps.hxx"
#include "build.hxx"

// The software was constructed for the ``beautifaction of solid
// models'' project at Cardiff University COMSC during 1999 and
// 2000. The first section of this file contains general comments and
// notes that could be useful if you are intending to look into the
// software internals in a serious manner. The actual code starts
// further down.
//
// Dear Reader,
//
// Due to circumstances this code was not completed to the degree that
// I myself would have prefered. However, as it stands is should
// perform robustly in the target area. No explict bug-list is given,
// but I have attempted to make the problems and limitations clear by
// the comments.  The code was not written to be efficient in time or
// memory resources.  Rather, it was intended to be fast and certain
// in the implementation, as well as being structured to make it
// understandable to a human. The ideas behind this code can be a
// little esoteric, so I have tried to structure the software to make
// the intention as clear as possible. If you are looking to modify
// the source, then please spend a little while to understand why I
// wrote a routine just-so. I have tried to provide all the
// information at the right points to enable you to work out what the
// original motivation was, which may mean that only a small change is
// needed even though at first it all looks rather incomprehensible.
//
// At the time of writing there is the problem that local variable
// structures can cause a stack overflow. So, if you find that the
// addition of a local variable, or an extra call or two, causes a
// previously running version of the program to do a core dump, then
// as a first thing, check the size of local variables, and maybe try
// declaring things as static to reduce the volume of material on the
// stack.
//
// In general, the abstract data types defined in this software are
// based on record structures of known and specific sizes. No use is
// made of linked lists for potential indefinite expansion of the
// input data size.  This is for two reasons. Firstly, I have yet to
// find a dynamic memory handler, in a procedural language, that can
// really stand up to the sort of thrashing it would take in a dynamic
// implementation of this sort of material. Secondly, some of the
// theoretical speed of the algorithms is based on certain operations
// being constant time. This only occurs because the information is
// stored in an array, and the array lookup is constant time. A link
// list implementation would tend to change this to linear time. In a
// frequently called central operation this can make a rather large
// difference to the overall time performance.
//
// Given more time I would write a memory handler specifically crafted
// for this application. Obviously, you can have a shot at it if you
// want. This would allow more dynamic data structures, and take away
// the hard coded upper bounds on the input data size. However, in
// practice, within the confines of this project so far, the input
// information has been limited to no more than 300 entities for a
// given solid. The memory requirements for processing this can be
// provided by a fairly typical desk top machine available at the
// beginning of the twenty-first century. The implementation can take
// a few hours to process the larger models, but I am told this is
// within tolerable levels. So, for pragmatic purposes of completing
// the algorithm design, and proof of concept, any further concern for
// the speed, memory, and input data size constraints does not seem
// justified. It is the method that is being developed for this
// project, not the actual software.
//
// One distinct problem with the code, which I will try to fix before
// I leave, but I don't know how far I will get, is that more use
// should be made of pointers to data structures being handed as
// arguments to other routines. At the moment there is a very heavy
// use of the program stack, and of copying of large data
// structures. This can cause stack crashes for around 200-300
// vertices, and of course makes the code take longer than it really
// needs to, due to all that copying back and forth of data.  It would
// be a good idea to convert code to use pointers to data structure
// where possible.  This is a straight forward, local, but tedious,
// process of converting pass by value to pass by reference, and
// making sure that the called routine does not mash the data
// structure. Doing this would mean that there would be no need to
// recompile the code with smaller data structures, to get it to run
// faster, as I have sometimes done.
//
// Another matter that is more algorithmic is to implement more
// sophisticated factoring algorithms. This involves a bit of thought
// and a few decisions, in particular the routine(s)
// partition::factor() that use information about a partitioning of
// the collection of vertices to factor the not just the pointset, but
// also the rest of the boundary representation. As it stands a fairly
// naive approach is used which can have the effect that the algorithm
// does not determine all appropriate symmetries.  However, to get
// around this the full_group structure is returned (see comments
// about find_factored_regularities) which gives the less
// sophisticated, but more reliable, information, in this case.
//
// I strongly recomend that, if you are going to modify this code, you
// start by going through it, carefully bit by bit, adjust it so that
// most of the variable passing of the large structures is by
// reference, instead of by value. If done with honesty, this should
// give you a good feel for some of the structure of the code. As well
// it should make it run a few times faster. At the moment, increasing
// the size of the basic data structures, even if the extra memory is
// not used, slows it down a lot. So, I expect a good pragmatic
// speed-up from a pointer based approach.
//
// If you want a quick idea of how to use/call the code then check out
// the routines `main' (obviously), `find_regularity', and
// `find_factored_regularities'. The major function of main is to
// parse and interpret the command line, there are a fair few flags
// and file formats that it accepts. The find_regularity routine takes
// a point collection, and determines the collection of permutations
// that correspond to the approximate symmetry,
// find_factored_regularities looks for partitions of the collection
// of points and then calls find_regularity to determine the
// approximate symmetry.
//
// If you are looking for some work to do, you could also make more of
// the members of the classes private in order to make the program
// more modular.  This would be a good idea if extensive use is going
// to be made of the types defined in here. However, you would need to
// go through and check all the uses, and declare a few classes to be
// C++ friends of some of the others. This is not entirely straight
// forward, but it is highly recomended if you want to get the flavour
// of the code. Once you have done this you should have a good idea of
// where every thing is.
//
// CAVEAT: I know it can be done, because I have in mind how to do it,
// but, a lot of that information pertains to my particular experience
// and training. If you are mathematically inclined, then keep in mind
// that I have a PhD in the geometry and topology of algebraic
// structures.  Keep this flavour in mind, and it should, in a
// non-specific manner, help you to understand what I was working
// towards. If you are in to cryptic clues, I have one that may help,
// read the book `space paw'.  (by Gordon R Dickson 1969).
//
// Good luck,
//
// Regards,
//
// Bruce Mills.
//
//
// Determination of approximate partial symmetry of a point set.
// -------------------------------------------------------------
//
// Code and theory developed by 
// Bruce Mills (1999-2000).
//
// In terms of both principle and implimentation this file should be
// fairly self contained. Some basic use is made of common C++
// libraries, and reference is made to theory that can be followed up
// in the literature, but you should find that most of what you need
// to understand and run this code is already here.
//
// The routines in this file are aimed at solving a particular problem
// in computational geometry. Given a list of points find any
// reasonably justified approximate partial symmetries. Where,
// intuitively, an approximate symmetry is an isometry which moves the
// collection of points almost into coincidence with itself, and a
// partial symmetry is an approximate symmetry of a subset of the
// collection. Obviously, if the level of approximation is
// particularly large, and/or the subset is particularly small, (in
// number or dimension), then just about any isometry can be an
// approximate partial symmetry. Thus a large part of the development
// is in determining what justification there might be for selecting
// one and not another.
//
// I won't attempt any strong meta justification of my approach.
// Presumably you are reading this file because you have taken over
// after I left, and are trying to get to grips with my work.  You
// should have a copy of the paper ``approximate symmetry detection
// for reverse engineering''. Either there is a copy of it on your
// desk (my old desk), or you can get one from Frank, Ralph or
// Dave. In fact, there is a reasonable chance that you might even be
// Frank, in which case that solves the whole problem of explaining my
// algorithmic intentions. If you are not Frank, then tell him to give
// you an explanation of the basic approach. He should be able to give
// you the flavour.
//
// However, I will give here a broad description of what I am doing,
// in order to get you on the same wavelength. I will also include a
// basic explanation later of how this all relates to the full
// pragmatic problem of dealing with the data from a boundary
// representation model of a solid. But for most of these notes I will
// just talk about collections of points.
//
// For a concrete model think in terms of a collection of points in
// standard three dimensional space. A lot of this material also works
// for an arbitrary metric space, but I won't follow up on that point
// in these notes.
//
// What is the subject matter?
// ---------------------------
//
// Imagine a collection of points, P={p_i : i in [1..n]}. These might,
// for example, be the corners of a cube. An exact symmetry of these
// points is an isometry that maps the set P onto itself.  (An
// isometry is just a rigid motion and/or reflection). Ie, if T is an
// isometry, then if the set {T(p_i) : i in [1..n]} is just P all over
// again, then T is a symmetry of P, and we write this as T(P)=P. If
// the points are the corners of an axis-aligned and origin-centred
// cube, then any rotation by a multiple of 1/4 turns about an axis is
// such an isometry. However, if we ever-so-slightly move the points
// in irregular directions, then this correspondence does not
// occur. We could, and will, argue that the points are almost mapped
// to each other, however we then need an extra piece of
// information. We need to know which point is matched with which.
// Ie, in order to state that an isometry T maps P approximately onto
// P, we need a permutation f of [1..n] such that T(p_i) ~ p_f(i).
// (using ~ to mean approximately equal). As long as the points are
// distinct, and the symmetry exact, the permutation was obvious,
// because point p_i was mapped exactly onto of point p_f(i). But,
// once an approximation is allowed it is possible that there is no
// obvious mapping involved. This is especially true in computer based
// geometry, since the computer can't see by inspection what the
// permutation is even in the easy cases for a human.
//
// What is an approximate symmetry?
// --------------------------------
//
// Having said all of that, an isometry T is an approximate symmetry
// of a collection of n points P, to a level of tolerance t, if and
// only if there exists a permutation f of [1..n] such that
// ||T(p_i)-p_f(i)||<t.  That is, T is an approximate symmetry, if we
// can consistently state that it moves each point to within a
// distance t, of a nominal target point.  In order for this to be
// justified as an approximate symmetry, I require that the mapping f
// be unique. Ie, there is one and only one way to permute the points
// so that T is seen as an approximate symmetry. In the context of the
// algorithm, this is done by requireing that t is less than half the
// minimum distance between two points. Technically there may be valid
// cases for which this condition is violated, however it seems that
// for fixing up solid models those cases are mathematical
// technicalities rather than useful approximate symmetries. I make
// this comment in the context of the point set grouping, which helps
// to get rid of these types of awkward cases.
//
// What is a factored symmetry?
// ----------------------------
//
// I have already mentioned the idea that points should be distinct.
// In particular, if two of the points on the input list are the same
// then they should be replaced by a single copy. However, rather than
// being identicle, two points might be much closer to each other than
// to any other points. In this case we could argue that, looking at
// it through fuzzy glasses, they are the same point, and should be
// replaced with a single copy. The idea of a valid factoring, or
// grouping of the points is that we collect together points that are
// closer to each other than the tolerance level, and replace them by
// the centroid of the group. However, the idea of approximately equal
// can't always be used to group points consistently, so we look
// specifically for levels of tolerance for which the grouping is
// consistent, and see these as having special relation to the object
// in question. If such a factored set of points has a symmetry, then
// this is a factored symmetry of the whole object. (strictly a
// symmetry of the factored object).
//
// What is a partial symmetry?
// ---------------------------
//
// As previously mentioned a partial symmetry in the most general
// sense is a symmetry of a subset of the object. This is a simple,
// natural definition, but it includes too many things. So I restrict
// what is acceptable to, for a start, being a group. Ie, we need to
// find a group of symmetries that is a partial symmetry of a subset
// of the points which are not degenerate. The next requirement is
// that the group of symmetries should be a maximal group (ie no
// strict extension group is a symmetry of the same subset) and that
// the subset is also maximal (ie that no strict supeset of the subset
// has the same group of symmetries. I suggest that examination of the
// maximal subsets would lead to natural decompositions of the object
// into symmetric sub-assemblies, plus asymmetric whatever is left
// over.
//
// Where are all the isometries?
// -----------------------------
//
// You may notice that there is very little explicit use of isometries
// in this code. The reason is that they are computed with implicitly
// using permutations. We are used to the idea, when symmetries are
// exact, of having a finite collection of symmetries (plus maybe
// continuous rotation about a unique axis). However, if we allow a
// level of tolerance, the number of actual isometries leaps to
// infinity. Of course, a human being can distinguish that typically
// these fall into equivalence classes, and instead of including
// rotation by 90.000, 90.001, 90.002 and so on degrees, the person
// just says, rotation by about 90 degrees. The computer needs a way
// of doing this. The manner that I use is that two isometries are
// equivalent if they induce the same permutation, in the manner
// indicated above. But, if this is the case, then the composition of
// two isometries in a group will induce the composition of the two
// permutations (or it's not a valid, consistent symmetry group).  So,
// instead this code just uses the permutations. These can be
// identified as the permutations that preserve the distance between
// all the pairs of points. In essence the information about the
// object has been reduced to trigonometric information about the
// edges of triangles, which is a representation that is completely
// independent of the location and orientation of the object. It makes
// for the posibility of computation with approximate isometries, in
// which the computation is exact rather than involving complicated
// aproximate equality testing, and constantly re-deriving the same
// equivalence class via a different isometry.

///////////////////////////////////////////////////////////////////////////
//                                                                       //
//     *     *       * * *    * * *   * * *   * * *  * * * * *  *     *  //
//    * *    *      *     *  *     *  *    *    *        *      *     *  //
//   *   *   *      *     *  *     *  *    *    *        *      *     *  //
//  * * * *  *      *        *     *  * * *     *        *      * * * *  //
//  *     *  *      *   ***  *     *  *   *     *        *      *     *  //
//  *     *  *      *     *  *     *  *    *    *        *      *     *  //
//  *     *  * * *   * * *    * * *   *     * * * *      *      *     *  //
//                                                                       //
///////////////////////////////////////////////////////////////////////////

// This is a major break point in the code. Up until this point the
// routines have basically been involved with the construction of the
// underlying abstract data types on which the algorithm is to be
// based. Obviously the actual implimentation of the data types is
// needed for the implimentation of the algorithm. But, for
// modularity, they should be kept as distinct and self contained as
// possible. The code that follows this point is specifically the
// implimentation of the algorithms for detection of symmetry in
// collections of points.

// =================== Look for Approximate Regularities =============


// This checks that the injection preserves lengths to within the
// given tolerance. Just does it by checking order n^2
// pairs. (Avoiding duplicates).
//
// The symmetry detection part of the algorithm expressed in this file
// is simply the problem of finding permutations that preserve lengths
// to within the given tolerance. In this context, the use of the
// following routine is something of an inefficiency. The permutations
// are incremented through a predefined sequence of permutations, see
// the routine injection::next(), and then checked to see if they
// approximately preserve the length. However, most of the time when
// we go to the next permutation, most of the permutation is
// unaffected.  In these cases we could impliment a local
// check-on-update to the next() function. This would be order n
// instead of order n^2.
//
// Whether this is worth it though is another question, since the
// actual algorithm only goes to depth 4 in the search, and then uses
// a different tactic. But the above thought is relevant in particular
// when the ideas are applied to points on the surface of a sphere (no
// centroid) and anyway, would be an efficiency if you are looking to
// improve the pragmatic performance of the implimentation of the
// algorithm.

bool check_regularity (injection *m, pointset *p)
{
  if (m->n != p->n)
    return false;

  for (word i = 0; i < p->n; ++i) {
    for (word j = i + 1; j < p->n; ++j) {
      if (fabs ((p->item[i] - p->item[j]) . length() -
		(p->item[m->forward[i]] - p->item[m->forward[j]]) . 
		length()) >= p->epsilon) {
	return false;
      }
    }
  }
  return true;
}

// Given a collection of n points in m-space, and another point x,
// we can determine the distance of x from each of these n points.
// In this way we can build up a co-ordinate system m-space, in 
// which the ordinate values are always non negative. If the n points
// form an m-simplex, ie if there are m+1 of them, and they are in
// a non degenerate arrangement, then this actually forms a complete
// co-ordinate system, in which each point in m-space is given a
// unique collection of m+1 ordinates. 
//
// The following routine simply checks to see if the co-ordinates 
// of the first point x1, wrt the set m1, is roughly the same as
// the co-ordinates of x2, wrt the set m2.

// This computes the coordinates of p in the frame based on the
// distances from the points in the pointset.
//
// By frame co-ordinates I mean the set of distances from the
// elements of the pointset. For example, given a triangle in
// a plane, the location of a point in the plane is completely 
// specified once you know what the distance of the point is
// from each corner of the triangle. In three dimensions you
// need the a tetrahedron (four non degenerate points) to 
// do the same thing.
//
// So ... keep in mind that if the points in your data are
// three dimensional, then this routine will be generating,
// at times, four dimensional points (ie points with four
// co-ordinate values stored in them).

// c is the centroid for the registration set, find representation
// in base created by pointset and c. Using the distances from the
// point causes trouble (not unique if using the tolerances)

// Given two poinsets m1 and m2, implicitely determining a two
// co-ordinate systems, by using the metric, we want to find
// out if we can match the points in p1 and the points in p2
// according to the respective co-ordinate systems. The intention
// is that m1 and m2 are registration sets for p1 and p2, and that
// m1 and m2 are congruent. Then we want to know if p1 and p2 are
// congruent, and to find a specific injection that realises this
// congruency. All of this is done using the approximate metric.
//
// At the moment this uses an n^2 algorithm. If we used a data base
// lookup routine, for finding points inside rectangles in multi
// dimensional spaces, then it could be reduced to n log^4 n. 
// Whether this is worth it given the overhead is another question.

injection *create_permutation (pointset *p, word *map_idx, point const &c)
{
  injection *w = new injection (p->n);
  w->identity ();

  // Finding the closest doesn't really solve the problem
  for (word i = 0; i < p->n; ++i) {

    // Compute distances for point from registration set
    point c1 (p->order + 1);
    for (word l = 0; l < p->order; ++l)
      c1.coord[l] = (p->item[i] - p->item[p->reg_idx[l]]) . length ();
    c1.coord[p->order] = (p->item[i] - c) . length ();

    word reg = 0;
    // Compute disatnces for point from targets of registration set
    point c2 (p->order + 1);
    for (word l = 0; l < p->order; ++l)
      c2.coord[l] = (p->item[0] - p->item[map_idx[l]]) . length ();
    c2.coord[p->order] = (p->item[0] - c) . length ();

    numeric dist = (c1 - c2) . length ();
    for (word j = 1; j < p->n; ++j) {

      for (word l = 0; l < p->order; ++l)
	c2.coord[l] = (p->item[j] - p->item[map_idx[l]]) . length ();
      c2.coord[p->order] = (p->item[j] - c) . length ();

      numeric d = (c1 - c2) . length ();
      if (d < dist) {
	dist = d;
	reg = j;
      }

    }

    // Add closest to permutation (more secure than first closer than
    // p->epsilon, at least seems to avoid some trouble?)
    if (dist < p->epsilon)
      w->addpair (i, reg);
  }


  return w;
}

// N^2 algorithm to check if the two (ordered) pointsets are 
// structurally the same.
//
// Check the paper I wrote in the third quarter of 200, mentioned
// at the top of this file. By using a data base lookup approach
// this might be reduced a bit, maybe n log^4 n. It would have
// big multiplicative constants though, so might not be worth it
// for just a few hundered points. 

bool congruent (pointset *p, word *map_idx, word n)
{
  // pointset p is first point set, we only take the reg_idx
  // registration set points. Similarly for the target points in map_idx.

  // Last set in pointset 2 is new (pointset1 is fixed set from
  // registration set). Only compare the distances to this last point
  // from the other points in ps2 with those they are mapped to in
  // ps1.

  for (word j = 0; j < n; ++j) {
    numeric t = fabs ((p->item[p->reg_idx[n - 1]] - p->item[p->reg_idx[j]]) .
		      length() - 
		      (p->item[map_idx[n - 1]] - p->item[map_idx[j]])
		      . length());
    if (t >= p->epsilon)
      return false;
  }

  return true;
}


///////////////////////////////////////////////////////////////////////


// Determine the regularities of a collection of points.
//
// In this context a regularity is a permutation of the points
// that preserves the distance between pairs of points to within
// the required accuracy. This is refered to simply as a metric
// preserving permutation.
//
// The first argument is the collection of points for which the
// regularities should be determined. The second argument must
// be a subset of the first argument. This subset is known as
// a registration subset. The idea is that if you know the location
// of each of the points in the registration subset, then you can
// determine the location of all the points in the collection by
// making sure that all the distance match correctly.
//
// As an example, any non degenerate triangle taken from a set of
// points embedded in a Euclidean plane is suitable. If we know
// where an isometry takes that triangle, then we know where the
// isometry takes the whole pointset. In the plane we do not need
// more than 3 points in the registration subset, in space we may
// need up to 4 points, but no more. However the following routine
// does not actually use any information about the dimension of 
// the space. One effect of this is that points in space that
// happen to lie in a plane can be naturally handled as a two
// dimensional problem, rather than as a three dimensional 
// problem (the latter takes longer).
//
// The find_regularities routine takes information about
// a construction of vertices, edges and faces. In particular the
// vertices are numbered 0 .. n-1, where n is the number of vertices.
// The edges and faces are similarly numbered, starting at 0. This
// routine is given a list of points in space, being the location
// of the respective vertices. It is also given, in terms of the
// vertex, edge and face indices, information about which two 
// vertices each edge goes between, and which two faces each edge
// joins by virtue of being a common boundary. It also uses the
// implied tolerance, which is a bit less than half the shortest
// distance between two any two given vertices.
//
// The routine computes a number of collections of permutations.
// In principle these collections should be groups, however no
// reliance is made on this notion, nor is any specific attempt
// made to make sure it is so. The four collections are as follows:
//
// full_group: the complete set of permutations of the vertex indices
//             which preserve the distance between all points to within
//             the implied tolerance.
//
// node_group: Those permutations of the vertices which induce a
//             consistent permutation on both the edges and faces.
//
// edge_group: The permutation of the edges corresponding to the
//             node_group. This has exactly the same number of
//             elements as the node group.
//
// face_group: The permutation of the edges corresponding to the
//             node_group. This has exactly the same number of
//             elements as the node group.
//
// In principle, the node group is a subgroup of the full group,
// and the edge and face groups have the same number of elements
// as the full group. Both the full group and the node group 
// should in principle contain at least the identity permutation.

void find_regularity (pointset *p, dualgraph *g, realisation *r,
		      pseudosym *node_group, pseudosym *edge_group,
		      pseudosym *face_group, pseudosym *full_group,
		      point const &c)
{
  // Note that the registration set is the first p.order points in p
  // plus the centroid now!

  // This routine matches the registration set to the other points on
  // the assumption that the centroid will remain where it is.  Thus
  // each point is matched only to ones that are the same distance
  // from the centroid. This effectively puts the centroid into the
  // registration set, and reduces by 1 the number of points that are
  // required in the registration set. In principle this takes one
  // factor of n out of the time complexity of the central loop. So
  // find_frame actually builds a frame with the centroid in it, but
  // deletes the centroid and only returns the other points.

  // Sort the points in order of distance from the centroid.

  heap *h = new heap (p->n);
  for (word k = 0; k < p->n; ++k)
    h->push ((c - p->item[k]) . length (), k);

  // Keep order in separate array
  word *p_sort = new word[p->n];
  for (word k = 0; k < p->n; ++k)
    p_sort[k] = h->grab ();

  delete h;

  // The collection of points close to each specific registration point
  // now fall into a contiguous block. So for matching each registration
  // point we only need to vary the index within the corresponding block.
  // So compute the minimum and maximum index for each contiguous block.

  numeric dst[p->order];
  word min[p->order];
  word max[p->order];
  for (word j = 0; j < p->order; ++j) {
    dst[j] = (c - p->item[p->reg_idx[j]]) . length ();
    min[j] = p->n;
    max[j] = p->n;
  }

  for (word k = 0; k < p->n; ++k) {
    numeric t = (c - p->item[p_sort[k]]) . length ();
    for (word j = 0; j < p->order; ++j) {
      if (min[j] == p->n && t > dst[j] - p->epsilon)
	min[j] = k;
      if (max[j] == p->n && t > dst[j] + p->epsilon)
	max[j] = k;
    }
  }

  for (word j = 0; j < p->order; ++j) {
    if (max[j] == p->n)
      max[j] = p->n - 1;
  }

  // A precondition for the correct operation of this function is
  // that the registration simplex is a subset of the pointset.
  // So we look for all possible metric preserving permutations
  // of p by finding all the metric preserving injections of v
  // into p, and checking to see if they can be extended to make
  // a metric preserving permutation of p.
  //
  // This could in principle be done by trying all injections of 
  // v into p. However metric preservation is a monotonic property, 
  // if a partial injection of v into p does not preserve the metric, 
  // then neither does any extension of that injection. Thus a back
  // tracking tree scan has been used that extends a partial injection
  // only if it is metric preserving. If a full injection is constructed
  // then this is checked to see if it can be extended to a permutation.


  // indices of the central multi-nested for-loop
  word i[p->order];
  i[0] = min[0];

  // Indices of pointset in p tried to match with registration set
  word map_idx[p->order];
  map_idx[0] = p_sort[i[0]];

  pointset m (p->order, p->item[0].d); 
  m.item[0] = p->item[p_sort[i[0]]];
  m.epsilon = p->epsilon;

  int index = 0;

  while (index >= 0) {

    if (congruent (p, map_idx, index + 1)) {
      if (index + 1 == p->order) {
	// Maximal search depth reach, check if rest matches
	injection *node_perm = create_permutation (p, map_idx, c);
	if (check_regularity (node_perm, p)) {
	  // Permutation allowed.
	  full_group->add_permutation (node_perm);
	  if (r->scheme.def) {
	    // check if it is consistent with the scheme as well
	    injection *edge_perm = new injection (g->n);
	    edge_perm->identity ();
	    injection *face_perm = new injection (g->n);
	    face_perm->identity ();
	    if (node_perm->cando (g, edge_perm, face_perm)) { 
	      if (face_perm->cando (r, g->minface(), g->maxface())) {
		node_group->add_permutation (new injection (node_perm));
		edge_group->add_permutation (edge_perm);
		face_group->add_permutation (face_perm);
	      } else {
		delete edge_perm;
		delete face_perm;
	      }
	    } else {
	      delete edge_perm;
	      delete face_perm;
	    }
	  }
	} else {
	  delete node_perm;
	}
	// multi-nest for loop next
	i[index]++;
	while (i[index] > max[index] && index >= 0) {
	  --index;
	  ++i[index];
	}
	if (index >= 0)
	  map_idx[index] = p_sort[i[index]];
      } else {
	// Search depth not yet reached, but everything matches so far
	index++;
	i[index] = min[index];
	map_idx[index] = p_sort[i[index]];
      }
    } else {
      // Did not match, next option (multi-nested for loop next)
      i[index]++;
      while (i[index] > max[index] && index >= 0) {
	--index; 
	++i[index];
      }
      if (index >= 0) 
	map_idx[index] = p_sort[i[index]];
    }
    
  }

  delete p_sort;

  return;
}

/////////////////////////////////////////////////////////////////////////////

void find_frame (pointset *p, point const &c)
{
  // At this point we want to make a simplicial frame that is physcially
  // related to the object in question. The simplest way to do this is
  // to pick some of the points of the object itself. However using the
  // centroid as one of the points in the frame may be useful. But this
  // would need to be handled in the arguments somehow. Ie, in a sense
  // the way in which the frame is computed has to be either implicitly
  // or explicitly co-ordinated between the various routines.
  
  // note: this routine actually generates a frame which includes the
  // centroid of the set, and then deletes the centroid from the frame.
  // Since the find_regularity routine only attempts to match elements
  // of the frame with points that are the same distance from the 
  // centroid, the result is that it is effectively matching the frame
  // with the centroid, but always mapping the centroid to itself.
  // Which is the required result.

  if (p->reg_idx)
    delete[] p->reg_idx;
  p->reg_idx = new word[p->item[0].d];

  // Find the point furthest from the centroid
  word pl = 0;
  numeric d = (c - p->item[0]) . length ();
  word l = 1;
  while (l < p->n) {
    numeric dd = (c - p->item[l]) . length ();
    if (dd > d) {
      d = dd;
      pl = l;
    }
    ++l;
  }

  // Remember first registration point
  p->reg_idx[0] = pl;

  word num_pts = 1;

  // Build up as large a non degenerate simplex as is possible
  // using the points from the given pointset.
  for (word l = 1; l < p->item[0].d && l < p->n; ++l) {

    // Find next point maximizing measure of registration set
    pl = l;
    numeric m = p->measure (l, pl, c);
    word k = l + 1;
    while (k < p->n) {
      numeric mm = p->measure (l, k, c);
      if (mm > m) {
	m = mm;
	pl = k;
      }
      ++k;
    }

    // Stop if measure of registration set is too small
    if (m < epsilon)
      break;

    // Check if registration set is large enough (i.e. nothing lower dim.
    // embedded in higher dimensions).  
    
    // Vector from centroid to new position
    point qq = p->item[pl] - c;
    if (l == 1) {
      // Area, i.e distance between line and point
      point p1 = p->item[p->reg_idx[0]] - c;
      numeric d = (qq % p1) / p1.length ();
      numeric x = qq.length ();
      d = sqrt (x * x - d * d);
      if (d < p->epsilon)
	break;
    } else if (p->item[0].d == 3) {
      // Volume, i.e. distance between plane and point
      point p1 = p->item[p->reg_idx[0]] - c;
      point p2 = p->item[p->reg_idx[1]] - c;
      point pp (3);
      pp.coord[0] = p1.coord[1] * p2.coord[2] - p1.coord[2] * p2.coord[1];
      pp.coord[1] = p1.coord[2] * p2.coord[0] - p1.coord[0] * p2.coord[2];
      pp.coord[2] = p1.coord[0] * p2.coord[1] - p1.coord[1] * p2.coord[0];
      numeric x = fabs ((qq % pp) / pp.length ());
      if (x < p->epsilon)
	break;
    }
    // FIXME: other cases not implemented.

    // Exchange positions
    p->reg_idx[num_pts++] = pl;
  }

  p->order = num_pts;
  fprintf (stdout, "# The centroid is ");
  c.put ();
  fprintf (stdout, "\n# The %d-registration set is {", num_pts);
  for (word k = 0; k < p->order; ++k) {
    fprintf (stdout, "%u", p->reg_idx[k]);
    if (k + 1 < p->order)
      fprintf (stdout, ", ");
  }
  fprintf (stdout, "}\n");
  
  return;
}

// The main task of the routine find_factored_regularities is to determine
// appropriate ways of factoring the realisation it is given. The actual
// work of finding the regularites is done by find_regularities. The
// essential idea of factoring the realisation is to find a distance,
// called the tollerance, so that it is possible to assert the identity
// of two points that are closer together than this, in a consistant
// fashion. In general, the notion of approximate equality as described
// is not transitive, and so cannot be used consistently to group points.
// However, if we limit our attention to the specific points used in the
// realisation, then we might find that there are some tolerances that
// lead to consistent groupings, at least for those points in the model.
// These tolerances come in intervals. Two usable tolerances in the same
// interval have basically the same effect on the object. So the 
// find_factored_regularities routine determines implicitely the interval
// without explicitely computing a tolerance. It then works out how the
// points are to be collected together, and passes the resulting factored
// set of points on to the find_regularities routine. For further information
// about what the regularities are, see the comment to the find_regularities
// routine, as well as the information in the paper that I wrote, in the
// fourth quarter of 2000.

void find_factored_regularities (FILE *f, realisation *r, bool rectify,
				 bool first_only)
{
  fprintf (f, "###################### Start of symmetry detection #####################\n\n");
  r->put (f);

  // Initialize heap from pointset distances
  dualgraph g;
  g.complete (r->geometry.n);
  heap *h2 = new heap (g, r->geometry);

  // Create ordered distance/edge_id list from heap
  numeric *dist = new numeric[g.n];
  word *edge_id = new word[g.n];
  for(word l = 0; l < g.n; l++) {
    dist[l] = h2->key[0];
    edge_id[l] = h2->grab ();
  }

  // Heap no longer needed
  delete h2;

  // Create initial partition (each point separate)
  partition q (r->geometry.n);

  // Clustering loop
  bool first = true;
  word count_sets = 0;
  word l = 0;
  do {

    // Either first or l already points to next set, so distance is
    // between l and l-1! Even if we use the original pointset for the
    // first points make sure that the distances between the points is
    // larger than zero before doing anything.
    if ((first && dist[l] >= epsilon) ||
	(!first && dist[l] - dist[l - 1] >= epsilon && q.transitive ())) {

      first = false;

      // Found a transitive partition, store it.
      pointset *fac_p = q.factor (&r->geometry);
      dualgraph *fac_g = q.factor (&r->scheme);


      if (fac_p->n > 1) {

	++count_sets;

	fprintf (f,
		 "\n######################### Level %3d Symmetries #########################\n\n# The partition\n\n",
		 count_sets);
	q.put (f);
	
	// Check symmetry.
	fprintf (f, "\n# The factored point set\n\n");
	fac_p->put (f);
	fac_p->epsilon = fac_p->character () / 2.0;
	fprintf (f, "\n# The tolerance level is %4.2f\n", fac_p->epsilon);
	
	// Centroid
	point c = fac_p->centroid ();

	// Find registration set
	find_frame (fac_p, c);
	
	// Find symmetries
	pseudosym *full_group = new pseudosym (fac_p->n * fac_p->n);
	pseudosym *node_group = NULL;
	pseudosym *edge_group = NULL;
	pseudosym *face_group = NULL;
	if (r->scheme.def) {
	  node_group = new pseudosym (fac_p->n * fac_p->n);
	  edge_group = new pseudosym (fac_p->n * fac_p->n);
	  face_group = new pseudosym (fac_p->n * fac_p->n);
	}
	find_regularity (fac_p, fac_g, r, node_group, edge_group, face_group,
			 full_group, c);
	
	if (r->scheme.def) {
	  // Only report this if scheme defined in input set
	  fprintf (f, "\n# The vertex symmetry group\n\n");
	  node_group->put (f);
	  fprintf (f, "\n# The edge symmetry group\n\n");
	  edge_group->put (f);
	  fprintf (f, "\n# The face symmetry group\n\n");
	  face_group->put (f);
	}
	fprintf (f, "\n# The point set symmetry group\n\n");
	full_group->put (f);
	
	if (rectify && fac_p->order == 3) {
	  // FIXME: Rectification...
	  pointset ps2 (*fac_p);
	  pointset newps (*fac_p);
	  pointset *c_ps = &ps2;
	  pointset *n_ps = &newps;
	  word iter = 0;
	  while (iter < 500) {
	    if (node_group)
	      node_group->rectify (c_ps, n_ps, c);
	    else
	      full_group->rectify (c_ps, n_ps, c);
	    numeric d = 0.0;
	    for (word l = 0; l < c_ps->n; ++l)
	      d += (c_ps->item[l] - n_ps->item[l]) . length ();
	    d /= static_cast<numeric> (c_ps->n);
	    if (d < 10E-5)
	      break;
	    pointset *h = c_ps;
	    c_ps = n_ps;
	    n_ps = h;
	  }

	  fprintf (f, "\n# The rectified object\n");
	  ps2.put (f);
	  fprintf (f,
		   "\n# The edge length matrix of the rectified object\n#\n");
	  ps2.metric ();
	}
	
	delete full_group;
	if (r->scheme.def) {
	  delete node_group;
	  delete edge_group;
	  delete face_group;
	}

      } else {
	break; // Only one point in factored set, so stop
      }

      delete fac_p;
      delete fac_g;

      if (first_only)
	break;

    }

    // Do not try to merge last two classes as partition is ignored.
    if (l < g.n - 1) {
      // Merge next two classes
      q.merge (g.item[edge_id[l]].vertex[0], g.item[edge_id[l]].vertex[1]);
    }

    ++l;
  } while (l < g.n);

  // Last set is single point in any case, so we can ignore it
  
  delete dist;
  delete edge_id;

  fprintf (f,
	   "\n########## End of symmetry detection, no more valid partitions #########\n");
}
