// xfig.hxx - xfig records
//
// Copyright (C) 2000 Bruce I Mills <B.I.Mills@cs.cf.ac.uk>
// Copyright (C) 2001-2002 Frank C Langbein <frank@langbein.org>
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 675 Mass Ave, Cambridge, MA 02139, USA.

// Data types in this file:
//
// class xfig_point_record ... for handling xfig file records
// class xfig_line_record .... for handling xfig file records
// class xfig_text_record .... for handling xfig file records

// Indicates how big the field is in xfig coordinates when outputing a
// pointset as an xfig file. The xfig_deep indication is a parameter
// for a perspective projection, the final data output is two
// dimensional.  This is mainly for generating views so that the
// pointsets can be examined visualy.

#define xfig_wide 12000
#define xfig_high 12000
#define xfig_deep 12000

// The following just sets which xfig version we are using, I don't
// using anything that splits the difference, but a valid header is
// needed to make it all work.

// #define xfig_version "FIG 3.1"
#define xfig_version "FIG 3.2"

word xfig_header(FILE *f, char *ver);

class point;

// ----------------- xfig point record class ------------------------

// I have written up some material that allows xfig files to be used.
// The information available is two dimensional, so in a direct
// approach some of the information is lost. However, it does enable
// you to take a quick peek at what the output looks like. The
// resulting file is a perspective projection of the three dimensional
// model. Part of the information that is recorded is the co-ordinates
// of each point. So, in this sense the full information is
// recorded. It is also possible, as the code stands, to use xfig to
// produce an xfig file, and then read it back into freg. I had the
// intention of also allowing the placement of co-ordinates in the
// diagram, and then reading these back as well. This would enable a
// full skeleton to be defined in xfig. You might want to add this
// capacity, and then it is a lot easier to generate some simple test
// models. However, since you will also want to interface this to ACIS
// solid modeler routines, you may find it better just to extract the
// combinatorial information from an ACIS model which has been built
// using the ACIS library. It is a bit buggy, but you are supposed to
// be using it for this project anyway. So, that may be the best way
// to go. Depends on your own personal disposition.

// Write a small circle to represent a single point in an xfig file.
//
// xfig format for an ellipse:
//
// 1 3 0 1 0 7 100 0 -1 0.000 1 0.0000 .... some line stuff
// x-coord y-coord
// radius-1 radius-2
// x-coord y-coord
// x-coord y-coord

class xfig_point_record
{
public:
  int    field01;   // 3       (some sort of style file)
  int    field02;   // 0       line style (0=solid 2=dashed)
  int    field03;   // 1       line thickness
  int    field04;   // 0       pen colour
  int    field05;   // 7       fill colour
  int    field06;   // 100     xfig layer depth
  int    field07;   // 0       (doesnt seem to do anything)
  int    field08;   // -1      fill style (-1=none)
  float  field09;   // 0.000   gap to solid ratio
  int    field10;   // 1
  float  field11;   // 0.000
  int    field12;   //     x coord
  int    field13;   //     y coord
  int    field14;   //     radius
  int    field15;   //     radius
  int    field16;   //     x coord
  int    field17;   //     y coord
  int    field18;   //     x coord
  int    field19;   //     y coord  
  void get(FILE *f);
  word put(FILE *f);
  void standard(point p);
};

// ---------------- xfig line record class ------------------------

// The xfig line record is really a polyline. So the basic geometric 
// data is a sequence of points. A lot of the rest of the data is 
// standard line thickness, colour, fill style, line style and so
// on. With line style we have solid, dashed, dash-dot, dash-dot-dot,
// in each case other than solid, the ratio of the gap to dash is
// recorded.
//
// If you set the forward arrow flag, then another five fields
// (int int float float float) are inserted into the record.
// Ditto for the reverse arrow flag. (So don't do this).
//
// Not all of this stuff is implimented here, since what I mostly
// need for this is a single line segement, going between exactly
// two points. So the record length is set rather than variable.
// It might at some time be worth it to add other features, but
// for the moment that would not be particularly useful.

class xfig_line_record
{
public:

  int   field01;  // 1       (some sort of style file)
  int   field02;  // 0       line style (0=solid 2=dashed)
  int   field03;  // 1       line thickness
  int   field04;  // 0       pen colour
  int   field05;  // 7       fill colour
  int   field06;  // 100     xfig layer depth
  int   field07;  // 0       (doesnt seem to do anything)
  int   field08;  // -1      fill style (-1=none)
  float field09;  // 0.000   gap to solid ratio
  int   field10;  // 0       join style (0=miter 1=round)
  int   field11;  // 0       cap style (0=butt 1=round)
  int   field12;  // -1      fill intensity
  int   field13;  // 0       forward arrow flag
  int   field14;  // 0       back arrow flag
  int   field15;  // 2       number of points
  int   field16;  //         x coord of point 1
  int   field17;  //         y coord of point 1
  int   field18;  //         x coord of point 2
  int   field19;  //         y coord of point 2
  void get(FILE *f);
  word put(FILE *f);
  void standard(point p1, point p2); // set all fields to standard values
};


// ----------------- xfig text record class ------------------

// Write a text object record to an xfig file

class xfig_text_record
{
public:
  int   field01;
  int   field02;
  int   field03;
  int   field04;
  int   field05;
  int   field06;
  float field07;
  int   field08;
  int   field09;
  int   field10;
  int   field11;
  int   field12;
  char  field13[100];
  void get(FILE *f);
  word put(FILE *f);
  void standard(point p, char *s);
};
